import string
import sys

import hypothesis.strategies as st
import pytest
from fakeparser import Parser
from hypothesis import example
from hypothesis import given
from hypothesis import settings
from test_reader import get_entries_random
from test_reader import search_entries_random
from test_reader import with_call_entries_method
from test_types import test_highlighted_string_extract
from test_types import test_highlighted_string_roundtrip

from reader import make_reader
from reader.types import HighlightedString


# Skip, since these add about 10s to the test suite
# and fail intermittently because the deadline gets exceeded.
# Especially skip on PyPy, since they're even slower there,
# and also increase the likelyhood of tests failing due to sqlite3 brittleness.
#
# Things to do when we reconsider:
#
# * use hypothesis profiles
# * write better tests, so we can actually get value from this
#
# Existing issue: https://github.com/lemon24/reader/issues/188
#
pytestmark = pytest.mark.skip()


@st.composite
def data_and_kwargs(draw):
    data = draw(
        st.lists(
            st.tuples(st.integers(), st.integers(), st.datetimes(), st.datetimes())
        )
    )

    kwargs = draw(
        st.fixed_dictionaries(
            {},
            optional={
                'read': st.none() | st.booleans(),
                'important': st.none() | st.booleans(),
                'has_enclosures': st.none() | st.booleans(),
                'feed': st.none()
                | st.sampled_from([f'{t[0]}' for t in data] + [''])
                | st.text(),
                'entry': st.none()
                | st.sampled_from(
                    [(f'{t[0]}', f'{t[0]}, {t[1]}') for t in data] + [('', '')]
                )
                | st.tuples(st.text(), st.text()),
            },
        )
    )

    chunk_size = draw(st.integers(0, len(data) * 2))

    return data, kwargs, chunk_size


@pytest.mark.slow
@with_call_entries_method
@given(data_and_kwargs=data_and_kwargs())
@settings(deadline=1000 if sys.implementation.name == 'pypy' else 400)
def test_sort_and_filter_subset_basic(data_and_kwargs, pre_stuff, call_method):
    entry_data, kwargs, chunk_size = data_and_kwargs

    # can't use reader fixture because of
    # https://github.com/pytest-dev/pytest/issues/916
    # TODO: use the make_reader() fixture if possible
    reader = make_reader(':memory:')

    reader._storage.chunk_size = chunk_size

    parser = Parser()
    reader._parser = parser

    for feed_id, entry_id, feed_updated, entry_updated in entry_data:
        seen_feed = feed_id in parser.feeds
        feed = parser.feed(feed_id, feed_updated)
        parser.entry(feed_id, entry_id, entry_updated)
        if not seen_feed:
            reader.add_feed(feed.url)

    reader.update_feeds()
    pre_stuff(reader)

    expected = [
        (fid, eid) for fid, entries in parser.entries.items() for eid in entries
    ]

    actual = [eval(e.id) for e in call_method(reader)]

    if call_method not in (get_entries_random, search_entries_random):
        assert len(expected) == len(actual)
        assert set(expected) == set(actual)
    else:
        assert set(expected) >= set(actual)

    actual = [eval(e.id) for e in call_method(reader, **kwargs)]
    assert set(expected) >= set(actual)


@st.composite
def maybe_highlighted_words_and_markers(draw):
    random = draw(st.randoms(use_true_random=True))
    marker = ''.join(random.choices(string.ascii_letters, k=20))
    before = f'>{marker}>'
    after = f'<{marker}<'
    maybe_highlighted_words = st.lists(st.tuples(st.text(), st.booleans()))
    return draw(maybe_highlighted_words), before, after


@pytest.mark.slow
@given(maybe_highlighted_words_and_markers())
def test_highlighted_string_extract_fuzz(maybe_highlighted_words_and_markers):
    words, before, after = maybe_highlighted_words_and_markers
    input = ''.join(f'{before}{w}{after}' if h else w for w, h in words)
    value = ''.join(w for w, _ in words)
    highlights = [w for w, h in words if h]
    test_highlighted_string_extract(input, value, highlights, before, after)


@pytest.mark.slow
@given(st.lists(st.tuples(st.text(), st.booleans())), st.text(), st.text())
def test_highlighted_string_apply_fuzz(words, before, after):
    slices = []
    current_index = 0
    for word, is_highlight in words:
        next_index = current_index + len(word)
        if is_highlight:
            slices.append(slice(current_index, next_index))
        current_index = next_index

    string = HighlightedString(''.join(w for w, _ in words), slices)

    expected = ''.join(f'{before}{w}{after}' if h else w for w, h in words)

    assert string.apply(before, after) == expected


@pytest.mark.slow
@given(maybe_highlighted_words_and_markers())
def test_highlighted_string_roundtrip_fuzz(maybe_highlighted_words_and_markers):
    words, before, after = maybe_highlighted_words_and_markers
    input = ''.join(f'{before}{w}{after}' if h else w for w, h in words)
    test_highlighted_string_roundtrip(input, before, after)



from datetime import datetime
from unittest.mock import ANY
from unittest.mock import MagicMock

import bs4
import pytest
from fakeparser import Parser

from reader import Content
from reader import HighlightedString
from reader import InvalidSearchQueryError
from reader import SearchError
from reader import StorageError
from reader._search import Search
from reader._search import strip_html
from reader._sqlite_utils import DBError
from reader._sqlite_utils import require_version


def test_bs4_import_error(storage, monkeypatch):
    search = Search(storage)
    search.enable()
    monkeypatch.setattr('reader._search.bs4', None)
    monkeypatch.setattr('reader._search.bs4_import_error', ImportError('reason'))

    with pytest.raises(SearchError) as excinfo:
        search.update()
    assert 'search dependencies' in str(excinfo.value.message)
    assert isinstance(excinfo.value.__cause__, ImportError)


STRIP_HTML_DATA = [(i, i) for i in [None, 10, 11.2, b'aabb', b'aa<br>bb']] + [
    ('aabb', 'aabb'),
    ('aa<br>bb', 'aa\nbb'),
    ('aa<p>bb', 'aa\nbb'),
    ('<script>ss</script>bb', 'bb'),
    ('<noscript>ss</noscript>bb', 'bb'),
    ('<style>ss</style>bb', 'bb'),
    ('<title>ss</title>bb', 'bb'),
    ('aa<script>ss</script>bb', 'aa\nbb'),
    ('aa<noscript>ss</noscript>bb', 'aa\nbb'),
    ('aa<style>ss</style>bb', 'aa\nbb'),
    ('aa<title>tt</title>bb', 'aa\nbb'),
    ('<head><script>ss</script></head>bb', 'bb'),
    ('<head><noscript>ss</noscript>bb', 'bb'),
    ('<head><style>ss</style></head>bb', 'bb'),
    ('<head><title>tt</title>bb', 'bb'),
    ('<head>aa<script>ss</script>bb', 'aa\nbb'),
    ('<head>aa<noscript>ss</noscript></head>bb', 'aa\nbb'),
    ('<head>aa<style>ss</style>bb', 'aa\nbb'),
    ('<head>aa<title>tt</title></head>bb', 'aa\nbb'),
    (
        """
        <head>
            aa
            <title>tt</title>
            <p>bb
            <script>ss</script>
            <b>cc
            <noscript>nn</noscript>
            <style>ss</style>
            dd
        </head>
        ee
        """,
        'aa\nbb\ncc\ndd\nee',
    ),
]


@pytest.mark.parametrize('input, expected_output', STRIP_HTML_DATA)
# We test all bs4 parsers, since we don't know/care what the user has installed.
@pytest.mark.parametrize(
    'features',
    [
        None,
        # lxml fails to build on pypy, see the comments in setup.py for details.
        pytest.param(
            'lxml', marks=pytest.mark.skipif("sys.implementation.name == 'pypy'")
        ),
        'html.parser',
        'html5lib',
    ],
)
def test_strip_html(input, expected_output, features):
    output = strip_html(input, features)
    if isinstance(output, str):
        output = '\n'.join(output.split())

    # Special-case different <noscript> handling by html5lib.
    # https://www.crummy.com/software/BeautifulSoup/bs4/doc/#differences-between-parsers
    is_html5lib = any(
        [
            features == 'html5lib',
            features is None
            and 'html5lib' in type(bs4.BeautifulSoup('').builder).__module__,
        ]
    )
    if is_html5lib and isinstance(input, str) and '<noscript>' in input:
        assert '<noscript>' not in output
        return

    assert output == expected_output


def enable_search(storage, _, __):
    Search(storage).enable()


def disable_search(storage, _, __):
    Search(storage).disable()


def is_search_enabled(storage, _, __):
    Search(storage).is_enabled()


def update_search(storage, _, __):
    Search(storage).update()


def search_entries_chunk_size_0(storage, _, __):
    list(
        Search(storage).search_entries_page('entry', datetime(2010, 1, 1), chunk_size=0)
    )


def search_entries_chunk_size_1(storage, _, __):
    list(
        Search(storage).search_entries_page('entry', datetime(2010, 1, 1), chunk_size=1)
    )


def search_entry_counts(storage, _, __):
    Search(storage).search_entry_counts('entry', now=datetime(2010, 1, 1))


def search_entry_last(storage, feed, entry):
    Search(storage).search_entry_last('entry', (feed.url, entry.id))


@pytest.mark.slow
@pytest.mark.parametrize(
    'pre_stuff, do_stuff',
    [
        (None, enable_search),
        (None, disable_search),
        (None, is_search_enabled),
        (enable_search, update_search),
        (enable_search, search_entries_chunk_size_0),
        (enable_search, search_entries_chunk_size_1),
        (enable_search, search_entry_counts),
        (enable_search, search_entry_last),
    ],
)
def test_errors_locked(db_path, pre_stuff, do_stuff):
    """All methods should raise SearchError when the database is locked."""

    from test_storage import check_errors_locked

    check_errors_locked(db_path, pre_stuff, do_stuff, SearchError)


def enable_and_update_search(storage):
    search = Search(storage)
    search.enable()
    search.update()


def iter_search_entries_chunk_size_0(storage):
    return Search(storage).search_entries_page(
        'entry', datetime(2010, 1, 1), chunk_size=0
    )


def iter_search_entries_chunk_size_1(storage):
    return Search(storage).search_entries_page(
        'entry', datetime(2010, 1, 1), chunk_size=1
    )


def iter_search_entries_chunk_size_2(storage):
    return Search(storage).search_entries_page(
        'entry', datetime(2010, 1, 1), chunk_size=2
    )


def iter_search_entries_chunk_size_3(storage):
    return Search(storage).search_entries_page(
        'entry', datetime(2010, 1, 1), chunk_size=3
    )


@pytest.mark.slow
@pytest.mark.parametrize(
    'iter_stuff',
    [
        pytest.param(
            iter_search_entries_chunk_size_0,
            marks=pytest.mark.xfail(raises=StorageError, strict=True),
        ),
        iter_search_entries_chunk_size_1,
        iter_search_entries_chunk_size_2,
        iter_search_entries_chunk_size_3,
    ],
)
def test_iter_locked(db_path, iter_stuff):
    """Methods that return an iterable shouldn't block the underlying storage
    if the iterable is not consumed.

    """
    from test_storage import check_iter_locked

    check_iter_locked(db_path, enable_and_update_search, iter_stuff)


class ActuallyOK(Exception):
    pass


def call_search_entries(search, query):
    try:
        next(search.search_entries(query, datetime(2010, 1, 1)))
    except StopIteration:
        raise ActuallyOK


def call_search_entry_counts(search, query):
    search.search_entry_counts(query, datetime(2010, 1, 1))
    raise ActuallyOK


@pytest.mark.parametrize(
    'query, exc_type',
    [
        ('\x00', InvalidSearchQueryError),
        ('"', InvalidSearchQueryError),
        # For some reason, on CPython * works when the filtering is inside
        # the CTE (it didn't when it was outside), hence the ActuallyOK.
        # On PyPy 7.3.1 we still get a InvalidSearchQueryError.
        # We're fine as long as we don't get another exception.
        ('*', (ActuallyOK, InvalidSearchQueryError)),
        ('O:', InvalidSearchQueryError),
        ('*p', InvalidSearchQueryError),
    ],
)
@pytest.mark.parametrize(
    'call_method',
    [
        call_search_entries,
        call_search_entry_counts,
    ],
)
def test_invalid_search_query_error(storage, query, exc_type, call_method):
    # We're not testing this in test_reader_search.py because
    # the invalid query strings are search-provider-dependent.
    search = Search(storage)
    search.enable()
    with pytest.raises(exc_type) as excinfo:
        call_method(search, query)
    if isinstance(exc_type, tuple) and ActuallyOK in exc_type:
        return
    assert excinfo.value.message
    assert excinfo.value.__cause__ is None


def test_minimum_sqlite_version(storage, monkeypatch):
    mock = MagicMock(wraps=require_version, side_effect=DBError('version'))
    monkeypatch.setattr('reader._search.require_version', mock)

    search = Search(storage)
    search.enable()

    with pytest.raises(SearchError) as excinfo:
        search.update()
    assert 'version' in excinfo.value.message

    mock.assert_called_with(ANY, (3, 18))


# TODO: test FTS5 column names



import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Any
from typing import cast
from typing import Iterable
from typing import Optional
from typing import Tuple
from typing import Union

from ._types import EntryData
from ._types import EntryForUpdate
from ._types import EntryUpdateIntent
from ._types import FeedData
from ._types import FeedForUpdate
from ._types import FeedUpdateIntent
from ._types import ParsedFeed
from ._utils import PrefixLogger
from .exceptions import ParseError
from .types import ExceptionInfo

log = logging.getLogger("reader")

HASH_CHANGED_LIMIT = 24


def process_old_feed(feed: FeedForUpdate) -> FeedForUpdate:
    if feed.stale:
        # db_updated=None not tested (removing it causes no tests to fail).
        #
        # This only matters if last_updated is None *and* db_updated is
        # not None. The way the code is, this shouldn't be possible
        # (last_updated is always set if the feed was updated at least
        # once, unless the database predates last_updated).
        #
        feed = feed._replace(updated=None, http_etag=None, http_last_modified=None)
        log.info(
            "update feed %r: feed marked as stale, "
            "ignoring updated, http_etag and http_last_modified",
            feed.url,
        )
    return feed


def make_update_intents(
    old_feed: FeedForUpdate,
    now: datetime,
    global_now: datetime,
    parsed_feed: Union[ParsedFeed, ParseError],
    entry_pairs: Iterable[
        Tuple[EntryData[Optional[datetime]], Optional[EntryForUpdate]]
    ],
) -> Tuple[Optional[FeedUpdateIntent], Iterable[EntryUpdateIntent]]:
    updater = _Updater(
        old_feed,
        now,
        global_now,
        PrefixLogger(log, ["update feed %r" % old_feed.url]),
    )
    return updater.update(parsed_feed, entry_pairs)


@dataclass(frozen=True)
class _Updater:

    """This is an object only to make logging easier."""

    old_feed: FeedForUpdate
    now: datetime
    global_now: datetime
    log: Union[logging.Logger, logging.LoggerAdapter] = log

    def __post_init__(self) -> None:
        object.__setattr__(self, 'old_feed', process_old_feed(self.old_feed))

    @property
    def url(self) -> str:
        return self.old_feed.url

    @property
    def stale(self) -> bool:
        return self.old_feed.stale

    def should_update_feed(self, new: FeedData, entries_to_update: bool) -> bool:
        old = self.old_feed
        self.log.debug("old updated %s, new updated %s", old.updated, new.updated)

        if not old.last_updated:
            self.log.info("feed has no last_updated, treating as updated")
            assert not old.updated, "updated must be None if last_updated is None"
            return True

        if not new.updated:
            self.log.info("feed has no updated, treating as updated")
            return True

        if self.stale:
            # logging for stale happened in process_old_feed()
            return True

        # we only care if feed.updated changed if any entries changed:
        # https://github.com/lemon24/reader/issues/231#issuecomment-812601988
        #
        # for RSS, if there's no date element,
        # feedparser user lastBuildDate as .updated,
        # which may (obviously) change without the feed actually changing
        #
        if entries_to_update and (not old.updated or new.updated > old.updated):
            self.log.info("feed updated")
            return True

        # check if the feed content actually changed:
        # https://github.com/lemon24/reader/issues/179
        if not old.hash or new.hash != old.hash:
            self.log.debug("feed hash changed, treating as updated")
            return True

        # Some feeds have entries newer than the feed.
        # https://github.com/lemon24/reader/issues/76
        self.log.info("feed not updated, updating entries anyway")

        return False

    def compute_entry_updated(
        self, id: str, new: Optional[datetime], old: Optional[datetime]
    ) -> Optional[datetime]:
        def debug(msg: str, *args: Any) -> None:
            self.log.debug("entry %r: " + msg, id, *args)

        if not new:
            debug("has no updated, updating but not changing updated")
            debug("entry added/updated")
            return old or self.now

        if self.stale:
            debug("feed marked as stale, updating anyway")
            debug("entry added/updated")
            return new

        if old and new <= old:
            debug(
                "entry not updated, skipping (old updated %s, new updated %s)", old, new
            )
            return None

        debug("entry updated")
        return new

    def should_update_entry(
        self, new: EntryData[Optional[datetime]], old: Optional[EntryForUpdate]
    ) -> Tuple[Optional[EntryData[datetime]], bool]:

        updated = self.compute_entry_updated(new.id, new.updated, old and old.updated)

        if updated:
            return EntryData(**new._replace(updated=updated).__dict__), False

        # At this point, new should have .updated is not None.
        # otherwise, compute_entry_updated() would have returned
        # either old.updated or self.now.
        # Remove this when it stops doing that, as proposed in this comment:
        # https://github.com/lemon24/reader/issues/179#issuecomment-663840297
        assert new.updated is not None

        # If old is None, compute_entry_updated() returned something.
        assert old is not None

        # Check if the entry content actually changed:
        # https://github.com/lemon24/reader/issues/179
        #
        # We limit the number of updates due to only the hash changing
        # to prevent spurious updates for entries whose content changes
        # excessively (for example, because it includes the current time).
        # https://github.com/lemon24/reader/issues/225
        #
        if not old.hash or new.hash != old.hash:
            if (old.hash_changed or 0) < HASH_CHANGED_LIMIT:
                self.log.debug("entry %r: entry hash changed, updating", new.id)
                # mypy does not automatically "cast" new to EntryData[datetime]
                return cast(EntryData[datetime], new), True
            else:
                self.log.debug(
                    "entry %r: entry hash changed, "
                    "but exceeds the update limit (%i); skipping",
                    new.id,
                    HASH_CHANGED_LIMIT,
                )

        return None, False

    def get_entries_to_update(
        self,
        pairs: Iterable[Tuple[EntryData[Optional[datetime]], Optional[EntryForUpdate]]],
    ) -> Iterable[EntryUpdateIntent]:
        last_updated = self.now
        for feed_order, (new_entry, old_entry) in reversed(list(enumerate(pairs))):

            # This may fail if we ever implement changing the feed URL
            # in response to a permanent redirect.
            assert (
                new_entry.feed_url == self.url
            ), f'{new_entry.feed_url!r}, {self.url!r}'

            entry_new = not old_entry
            processed_new_entry, due_to_hash_changed = self.should_update_entry(
                new_entry, old_entry
            )

            if processed_new_entry:
                yield EntryUpdateIntent(
                    processed_new_entry,
                    last_updated,
                    self.global_now if entry_new else None,
                    feed_order,
                    0
                    if not due_to_hash_changed
                    else ((old_entry and old_entry.hash_changed or 0) + 1),
                )

    def get_feed_to_update(
        self,
        parsed_feed: ParsedFeed,
        entries_to_update: bool,
    ) -> Optional[FeedUpdateIntent]:
        if self.should_update_feed(parsed_feed.feed, entries_to_update):
            return FeedUpdateIntent(
                self.url,
                self.now,
                parsed_feed.feed,
                parsed_feed.http_etag,
                parsed_feed.http_last_modified,
            )
        if entries_to_update:
            return FeedUpdateIntent(self.url, self.now)
        return None

    def update(
        self,
        parsed_feed: Union[ParsedFeed, None, ParseError],
        entry_pairs: Iterable[
            Tuple[EntryData[Optional[datetime]], Optional[EntryForUpdate]]
        ],
    ) -> Tuple[Optional[FeedUpdateIntent], Iterable[EntryUpdateIntent]]:

        # Not modified.
        if not parsed_feed:

            # New feed shouldn't be considered new anymore.
            if not self.old_feed.last_updated:
                return FeedUpdateIntent(self.url, self.now), ()

            # Clear last_exception.
            if self.old_feed.last_exception:
                return FeedUpdateIntent(self.url, self.old_feed.last_updated), ()

            return None, ()

        if isinstance(parsed_feed, ParseError):
            exc_info = ExceptionInfo.from_exception(
                parsed_feed.__cause__ or parsed_feed
            )
            return FeedUpdateIntent(self.url, None, last_exception=exc_info), ()

        entries_to_update = list(self.get_entries_to_update(entry_pairs))
        feed_to_update = self.get_feed_to_update(parsed_feed, bool(entries_to_update))

        if not feed_to_update and self.old_feed.last_exception:
            # Clear last_exception.
            # TODO: Maybe be more explicit about this? (i.e. have a storage method for it)
            feed_to_update = FeedUpdateIntent(self.url, self.old_feed.last_updated)

        return feed_to_update, entries_to_update



"""
Homegrown SQL query builder.

Inspired by sqlitebuilder.mini;
https://sqlbuilder.readthedocs.io/en/latest/#short-manual-for-sqlbuilder-mini

See tests/test_sql_utils.py for some usage examples.

---

For a version of this that supports UNIONs see the second prototype in
https://github.com/lemon24/reader/issues/123#issuecomment-624045621

---

For a version of this that supports INSERT/UPDATE/DELETE, see
https://github.com/lemon24/reader/commit/7c97fccf61d16946176c2455be3634c5a8343e1b

The way things work now has changed slightly;
we'd probably make them flag keywords now.

To make VALUES bake in the parentheses, we just need to set:

    query.formats[0]['VALUES'] = '({value})'

That's to add one values tuple at a time. To add one column, we could do this:

* output keyword even if called with no args
* add(..., flag=...), and allow arbitrary flags;
  INSERT_INTO('x', 'y', flag='table') -> flag is "INTO table"
* parens_keywords = {'INSERT', 'VALUES'};
  different than subquery_keywords because it applies once to the whole set

---

To support marking arbitrary things as subqueries,
add a signalling tuple and a helper function:

    class _Subquery(tuple): pass

    def Subquery(*args) -> _Subquery:
        return _Subquery(args)  # from_arg() has to support 1-tuples

Then, in from_arg(), if arg is a _Subquery, set is_subquery.

Usage looks like this:

    Query().FROM(
        Subquery('alias', 'subquery'),
        ('alias', 'not subquery'),
    )

Alternatively, add an is_subquery kwarg to add():

    query = Query().FROM(('alias', 'subquery'), is_subquery=True)
    query.FROM(('alias', 'not subquery'))

---

To support using Queries as arguments directly,
without having to convert them to strings first,
allow _Thing.value to be a Query (and support it in);
then, in _lines_keyword(), convert queries to str and override is_subquery.

"""
import functools
import textwrap
from collections import defaultdict
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

_T = TypeVar('_T')
_U = TypeVar('_U')


_Q = TypeVar('_Q', bound='BaseQuery')
_QArg = Union[str, Tuple[str, ...]]


class _Thing(NamedTuple):
    value: str
    alias: str = ''
    keyword: str = ''
    is_subquery: bool = False

    @classmethod
    def from_arg(cls, arg: _QArg, **kwargs: Any) -> '_Thing':
        if isinstance(arg, str):
            alias, value = '', arg
        elif len(arg) == 2:
            alias, value = arg
        else:  # pragma: no cover
            raise ValueError(f"invalid arg: {arg!r}")
        return cls(_clean_up(value), _clean_up(alias), **kwargs)


class _FlagList(List[_T]):
    flag: str = ''


def _clean_up(thing: str) -> str:
    return textwrap.dedent(thing.rstrip()).strip()


class BaseQuery:

    keywords = [
        'WITH',
        'SELECT',
        'FROM',
        'WHERE',
        'GROUP BY',
        'HAVING',
        'ORDER BY',
        'LIMIT',
    ]

    separators: Mapping[str, str] = dict(WHERE='AND', HAVING='AND')
    default_separator = ','

    formats: Tuple[Mapping[str, str], ...] = (
        defaultdict(lambda: '{value}'),
        defaultdict(lambda: '{value} AS {alias}', WITH='{alias} AS {value}'),
    )

    subquery_keywords = {'WITH'}
    fake_keywords = dict(JOIN='FROM')
    flag_keywords = dict(SELECT={'DISTINCT', 'ALL'})

    def __init__(
        self,
        data: Optional[Mapping[str, Iterable[_QArg]]] = None,
        separators: Optional[Mapping[str, str]] = None,
    ) -> None:
        self.data: Mapping[str, _FlagList[_Thing]] = {}
        if data is None:
            data = dict.fromkeys(self.keywords, ())
        for keyword, args in data.items():
            self.data[keyword] = _FlagList()
            self.add(keyword, *args)

        if separators is not None:
            self.separators = separators

    def add(self: _Q, keyword: str, *args: _QArg) -> _Q:
        keyword, fake_keyword = self._resolve_fakes(keyword)
        keyword, flag = self._resolve_flags(keyword)
        target = self.data[keyword]

        if flag:
            if target.flag:  # pragma: no cover
                raise ValueError(f"{keyword} already has flag: {flag!r}")
            target.flag = flag

        kwargs: Dict[str, Any] = {}
        if fake_keyword:
            kwargs.update(keyword=fake_keyword)
        if keyword in self.subquery_keywords:
            kwargs.update(is_subquery=True)

        for arg in args:
            target.append(_Thing.from_arg(arg, **kwargs))

        return self

    def _resolve_fakes(self, keyword: str) -> Tuple[str, str]:
        for part, real in self.fake_keywords.items():
            if part in keyword:
                return real, keyword
        return keyword, ''

    def _resolve_flags(self, keyword: str) -> Tuple[str, str]:
        prefix, _, flag = keyword.partition(' ')
        if prefix in self.flag_keywords:
            if flag and flag not in self.flag_keywords[prefix]:
                raise ValueError(f"invalid flag for {prefix}: {flag!r}")
            return prefix, flag
        return keyword, ''

    def __getattr__(self: _Q, name: str) -> Callable[..., _Q]:
        # conveniently, avoids shadowing dunder methods (e.g. __deepcopy__)
        if not name.isupper():
            return getattr(super(), name)  # type: ignore
        return functools.partial(self.add, name.replace('_', ' '))

    def __str__(self) -> str:
        return ''.join(self._lines())

    def _lines(self) -> Iterable[str]:
        for keyword, things in self.data.items():
            if not things:
                continue

            if things.flag:
                yield f'{keyword} {things.flag}\n'
            else:
                yield f'{keyword}\n'

            grouped: Tuple[List[_Thing], ...] = ([], [])
            for thing in things:
                grouped[bool(thing.keyword)].append(thing)
            for group in grouped:
                yield from self._lines_keyword(keyword, group)

    def _lines_keyword(self, keyword: str, things: Sequence[_Thing]) -> Iterable[str]:
        for i, thing in enumerate(things):
            last = i + 1 == len(things)

            if thing.keyword:
                yield thing.keyword + '\n'

            format = self.formats[bool(thing.alias)][keyword]
            value = thing.value
            if thing.is_subquery:
                value = f'(\n{self._indent(value)}\n)'
            yield self._indent(format.format(value=value, alias=thing.alias))

            if not last and not thing.keyword:
                try:
                    yield ' ' + self.separators[keyword]
                except KeyError:
                    yield self.default_separator

            yield '\n'

    _indent = functools.partial(textwrap.indent, prefix='    ')


if TYPE_CHECKING:  # pragma: no cover
    _SWMBase = BaseQuery
else:
    _SWMBase = object

_SWM = TypeVar('_SWM', bound='ScrollingWindowMixin')


class ScrollingWindowMixin(_SWMBase):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.scrolling_window_order_by()

    def scrolling_window_order_by(
        self: _SWM, *things: str, desc: bool = False, keyword: str = 'WHERE'
    ) -> _SWM:
        self.__things = [_clean_up(t) for t in things]
        self.__desc = desc
        self.__keyword = keyword

        order = 'DESC' if desc else 'ASC'
        return self.ORDER_BY(*(f'{thing} {order}' for thing in things))

    def extract_last(self, result: Tuple[_T, ...]) -> Optional[Tuple[_T, ...]]:
        names = [t.alias or t.value for t in self.data['SELECT']]
        return tuple(result[names.index(t)] for t in self.__things) or None

    def add_last(self, last: Optional[Tuple[_T, ...]]) -> Sequence[Tuple[str, _T]]:
        self.__add_last()
        return self.__last_params(last)

    __make_label = 'last_{}'.format

    def __add_last(self: _SWM) -> None:
        op = '<' if self.__desc else '>'
        labels = (':' + self.__make_label(i) for i in range(len(self.__things)))
        comparison = BaseQuery({'(': self.__things, f') {op} (': labels, ')': ['']})
        self.add(self.__keyword, str(comparison).rstrip())

    def __last_params(self, last: Optional[Tuple[_T, ...]]) -> Sequence[Tuple[str, _T]]:
        return [(self.__make_label(i), t) for i, t in enumerate(last or ())]


class Query(ScrollingWindowMixin, BaseQuery):
    pass


if TYPE_CHECKING:  # pragma: no cover
    import sqlite3


def paginated_query(
    db: 'sqlite3.Connection',
    query: Query,
    params: Dict[str, Any] = {},  # noqa: B006
    chunk_size: Optional[int] = 0,
    last: Optional[_U] = None,
    row_factory: Optional[Callable[[Tuple[Any, ...]], _T]] = None,
) -> Iterator[Tuple[_T, _U]]:

    params = dict(params)

    if chunk_size:
        query.LIMIT(":chunk_size")
        params['chunk_size'] = chunk_size
    if last:
        params.update(query.add_last(last))  # type: ignore

    rv = (
        (row_factory(t) if row_factory else t, cast(_U, query.extract_last(t)))
        for t in db.execute(str(query), params)
    )

    # Consume the result to avoid blocking the database,
    # but only if the query is actually paginated
    # (we may need the pass-through for performance).
    if chunk_size:
        return iter(list(rv))

    return rv



"""
enclosure_tags
~~~~~~~~~~~~~~

Fix tags for MP3 enclosures (e.g. podcasts).

Adds a "with tags" link to a version of the file with tags set as follows:

* the entry title as title
* the feed title as album
* the entry/feed author as author

This plugin needs additional dependencies, use the ``unstable-plugins`` extra
to install them:

.. code-block:: bash

    pip install reader[unstable-plugins]

To load::

    READER_APP_PLUGIN='reader._plugins.enclosure_tags:init' \\
    python -m reader serve

Implemented for https://github.com/lemon24/reader/issues/50.
Became a plugin in https://github.com/lemon24/reader/issues/52.

"""
import tempfile
from urllib.parse import urlparse

import mutagen.mp3
import requests
from flask import Blueprint
from flask import request
from flask import Response
from flask import stream_with_context
from flask import url_for

blueprint = Blueprint('enclosure_tags', __name__)


ALL_TAGS = ('album', 'title', 'artist')
SET_ONLY_IF_MISSING_TAGS = {'artist'}


@blueprint.route('/enclosure-tags', defaults={'filename': None})
@blueprint.route('/enclosure-tags/<filename>')
def enclosure_tags(filename):
    def update_tags(file):
        emp3 = mutagen.mp3.EasyMP3(file)
        changed = False

        for key in ALL_TAGS:
            if key in SET_ONLY_IF_MISSING_TAGS and emp3.get(key):
                continue
            value = request.args.get(key)
            if not value:
                continue
            emp3[key] = [value]
            changed = True

        if changed:
            emp3.save(file)
        file.seek(0)

    def chunks(req):
        # Send the headers as soon as possible.
        # Some browsers wait for the headers before showing the "Save As" dialog.
        yield ''

        tmp = tempfile.TemporaryFile()
        for chunk in req.iter_content(chunk_size=2 ** 20):
            tmp.write(chunk)
        tmp.seek(0)

        update_tags(tmp)

        try:
            while True:
                data = tmp.read(2 ** 20)
                if not data:
                    break
                yield data
        finally:
            tmp.close()

    url = request.args['url']
    req = requests.get(url, stream=True)

    headers = {}
    for name in ('Content-Type', 'Content-Disposition'):
        if name in req.headers:
            headers[name] = req.headers[name]

    return Response(stream_with_context(chunks(req)), headers=headers)


def enclosure_tags_filter(enclosure, entry):
    filename = urlparse(enclosure.href).path.split('/')[-1]
    if not filename.endswith('.mp3'):
        return []

    args = {'url': enclosure.href, 'filename': filename}
    if entry.title:
        args['title'] = entry.title
    if entry.feed.title:
        args['album'] = entry.feed.title
    if entry.author or entry.feed.author:
        args['artist'] = entry.author or entry.feed.author

    return [('with tags', url_for('enclosure_tags.enclosure_tags', **args))]


def init(app):
    app.register_blueprint(blueprint)
    app.reader_additional_enclosure_links.append(enclosure_tags_filter)



"""
A thing to build APIs that work both with old-style forms and with JSON.

Contains no business logic.

See scripts/jscontrols.py for a minimal usage example.

"""
from urllib.parse import urljoin
from urllib.parse import urlparse

from flask import flash
from flask import get_flashed_messages
from flask import jsonify
from flask import redirect
from flask import request
from werkzeug.datastructures import MultiDict


def is_safe_url(target):
    ref_url = urlparse(request.host_url)
    test_url = urlparse(urljoin(request.host_url, target))
    return test_url.scheme in ('http', 'https') and ref_url.netloc == test_url.netloc


def redirect_to_referrer():
    if not request.referrer:
        return "no referrer", 400
    if not is_safe_url(request.referrer):
        return "bad referrer", 400
    return redirect(request.referrer)


def get_flashed_messages_by_prefix(*prefixes):
    messages = get_flashed_messages(with_categories=True)
    rv = []
    for pair in messages:
        category, message = pair
        if not isinstance(category, tuple):
            category = (category,)
        for prefix in prefixes:
            if not isinstance(prefix, tuple):
                prefix = (prefix,)
            category_prefix = category[: len(prefix)]
            if category_prefix == prefix:
                rv.append(message)
    return rv


class APIError(Exception):
    def __init__(self, message, category=None):
        super().__init__(message)
        self.message = message
        if category is not None:
            if not isinstance(category, tuple):
                category = (category,)
        self.category = category


class APIThing:
    def __init__(self, app_or_blueprint, rule, endpoint):
        self.actions = {}
        self.really = {}
        app_or_blueprint.add_url_rule(
            rule, endpoint, methods=['POST'], view_func=self.dispatch
        )
        (
            getattr(app_or_blueprint, 'add_app_template_global', None)
            or app_or_blueprint.add_template_global
        )(get_flashed_messages_by_prefix)

    def dispatch_form(self):
        action = request.form['action']
        func = self.actions.get(action)
        if func is None:
            return "unknown action", 400
        next = request.form.get('next-' + action)
        if next is None:
            next = request.form['next']
        if not is_safe_url(next):
            return "bad next", 400
        if self.really[func]:
            really = request.form.get('really-' + action)
            if really is None:
                really = request.form.get('really')
            target = request.form.get('target')
            if really != 'really':
                category = (action,)
                if target is not None:
                    category += (target,)
                flash("{}: really not checked".format(action), category)
                return redirect_to_referrer()
        try:
            rv = func(request.form)
            flash(rv)
        except APIError as e:
            category = (action,)
            if e.category:
                category += e.category
            flash("{}: {}".format(action, e), category)
            return redirect_to_referrer()
        return redirect(next)

    def dispatch_json(self):
        data = MultiDict(request.get_json())
        action = data['action']
        func = self.actions.get(action)
        if func is None:
            return "unknown action", 400

        try:
            rv = func(data)
            rv = {'ok': rv}
        except APIError as e:
            category = (action,)
            if e.category:
                category += e.category
            rv = {'err': str(e)}

        return jsonify(rv)

    def dispatch(self):
        if request.mimetype == 'application/x-www-form-urlencoded':
            return self.dispatch_form()
        if request.mimetype == 'application/json':
            return self.dispatch_json()
        return "bad content type", 400

    def __call__(self, func=None, *, really=False):
        def register(f):
            self.actions[f.__name__.replace('_', '-')] = f
            self.really[f] = really
            return f

        if func is None:
            return register
        return register(func)



"""
Config file support.

https://github.com/lemon24/reader/issues/177

"""
import copy
from collections.abc import Mapping
from dataclasses import dataclass
from dataclasses import field

from .plugins import _DEFAULT_PLUGINS
from .plugins import _PLUGINS
from reader import make_reader
from reader._plugins import Loader


MAKE_READER_IMPORT_KWARGS = ('storage_cls', 'search_cls')


def make_reader_from_config(*, plugins=None, plugin_loader=None, **kwargs):
    """Like reader.make_reader(), but:

    * If *_cls arguments are str, import them.
    * Load plugins.

    """
    loader = Loader()
    plugin_loader = plugin_loader or loader

    for name in MAKE_READER_IMPORT_KWARGS:
        # this is dead code until make_reader() actually gets those arguments
        thing = kwargs.get(name)
        if thing and isinstance(thing, str):
            # use the default loader for these (we always want exceptions)
            kwargs[name] = loader.load(name, wrap=True)

    plugins = plugins if plugins is not None else dict.fromkeys(_DEFAULT_PLUGINS)

    plugins_arg = kwargs['plugins'] = []
    for name in list(plugins):
        if name in _PLUGINS:
            plugins_arg.append(name)
        else:
            plugin = plugin_loader.load(name, wrap=True)
            if not plugin:
                continue
            plugins_arg.append(plugin)

    reader = make_reader(**kwargs)
    return reader


def make_reader_config(thing):
    if not isinstance(thing, Mapping):
        raise ValueError("config must be a mapping")
    # TODO: validate / raise nicer exceptions here
    return Config(
        copy.deepcopy(thing), sections={'cli', 'app'}, merge_keys={'reader', 'plugins'}
    )


def _merge_config(*configs, merge_keys=()):
    """Merge multiple make_app_from_config() kwargs dicts into a single one.

    plugins is assumed to be a dict and is merged. All other keys are replaced.

    """
    rv = {}
    to_merge = {}

    for config in configs:
        config = config.copy()
        for name in merge_keys:
            if name in config:
                to_merge.setdefault(name, []).append(config.pop(name))
        rv.update(config)

    for name, dicts in to_merge.items():
        rv[name] = _merge_config(*dicts, merge_keys=merge_keys)

    return rv


@dataclass
class Config:

    data: dict = field(default_factory=dict)
    sections: set = field(default_factory=set)
    merge_keys: set = field(default_factory=set)
    default_section: str = 'default'

    def __post_init__(self):
        self.sections.add(self.default_section)

        unknown_sections = self.data.keys() - self.sections

        if self.default_section in self.data:
            if unknown_sections:
                raise ValueError(f"unknown sections in config: {unknown_sections!r}")
        else:
            self.data[self.default_section] = {
                section: self.data.pop(section) for section in unknown_sections
            }

            # default is always first
            for section in list(self.data):
                if section != self.default_section:
                    self.data[section] = self.data.pop(section)

        for section in self.sections:
            self.data.setdefault(section, {})

    def merged(self, section, overrides=None):
        if section not in self.sections:
            raise ValueError(f"unknown section: {section!r}")

        return _merge_config(
            self.data[self.default_section],
            self.data[section],
            overrides or {},
            merge_keys=self.merge_keys,
        )

    def merge_all(self):
        config = copy.deepcopy(self)
        for section in list(config.data):
            if section != config.default_section:
                config.data[section] = config.merged(section)
        return config

    def make_reader(self, section, **kwargs):
        return make_reader_from_config(
            **self.merged(section, {'reader': kwargs}).get('reader', {}),
        )

    @property
    def all(self):
        return MultiMapping(list(self.data.values()))

    def __getitem__(self, key):
        return self.data[key]


@dataclass
class MultiMapping:

    mappings: list = field(default_factory=list)
    default_factory: callable = dict

    def __getitem__(self, key):
        return MultiMapping(
            [
                mapping.setdefault(key, self.default_factory())
                for mapping in self.mappings
            ],
            self.default_factory,
        )

    def __setitem__(self, key, value):
        for mapping in self.mappings:
            mapping[key] = value



from fakeparser import Parser
from utils import utc_datetime as datetime

from reader import Enclosure


def test_plugin(make_reader):
    reader = make_reader(':memory:', plugins=['reader.enclosure_dedupe'])
    reader._parser = parser = Parser()

    feed = parser.feed(1, datetime(2010, 1, 1))
    one = parser.entry(1, 1, datetime(2010, 1, 1))
    two = parser.entry(
        1,
        2,
        datetime(2010, 1, 1),
        enclosures=(Enclosure('href'), Enclosure('another one')),
    )
    three = parser.entry(
        1,
        3,
        datetime(2010, 1, 1),
        enclosures=(Enclosure('href', 'text', 1), Enclosure('href', 'json', 2)),
    )

    reader.add_feed(feed.url)
    reader.update_feeds()

    assert set((e.id, e.enclosures) for e in reader.get_entries()) == {
        (one.id, one.enclosures),
        (two.id, two.enclosures),
        (three.id, (Enclosure('href', 'text', 1),)),
    }



"""
Backport of pkgutil.resolve_name (added in Python 3.9).

Could have used something like the following::

    pkg_resources.EntryPoint.parse('none = ' + import_name).resolve()

but it requires setuptools to be installed.

TODO: Remove ._vendor.pkgutil when we drop Python 3.8.

"""
from .pkgutil import resolve_name

__all__ = ['resolve_name']



import pytest

from reader import ParseError


def test_fallback(requests_mock, make_reader):
    url = 'http://www.example.com/'

    reader = make_reader(':memory:', plugins=('reader.ua_fallback',))
    reader.add_feed(url)

    matcher = requests_mock.get(url, status_code=403)

    with pytest.raises(ParseError) as exc_info:
        reader.update_feed(url)

    assert '403' in str(exc_info.value)

    assert len(matcher.request_history) == 2
    first_ua, second_ua = [r.headers['User-Agent'] for r in matcher.request_history]

    assert first_ua.startswith('python-reader/')
    assert second_ua.startswith('feedparser/')
    assert second_ua.endswith(first_ua)


def test_noop(requests_mock, make_reader):
    url = 'http://www.example.com/'

    reader = make_reader(':memory:', plugins=('reader.ua_fallback',))
    reader.add_feed(url)

    matcher = requests_mock.get(url, status_code=404)

    with pytest.raises(ParseError) as exc_info:
        reader.update_feed(url)

    assert '404' in str(exc_info.value)
    assert len(matcher.request_history) == 1



import cProfile
import inspect
import math
import os.path
import pstats
import random
import sqlite3
import statistics
import sys
import tempfile
import timeit
from collections import OrderedDict
from contextlib import contextmanager
from contextlib import ExitStack
from datetime import datetime
from datetime import timedelta
from fnmatch import fnmatchcase
from functools import partial

import click
from jinja2.utils import generate_lorem_ipsum

root_dir = os.path.dirname(__file__)
sys.path.insert(0, os.path.join(root_dir, '../src'))
sys.path.insert(0, os.path.join(root_dir, '../tests'))

from fakeparser import Parser

from reader import make_reader
from reader._app import create_app, get_reader
from reader._config import make_reader_config


def get_params(fn):
    rv = []
    for param in inspect.signature(fn).parameters.values():
        assert (
            param.kind == param.POSITIONAL_OR_KEYWORD
        ), f"parameter {param.name} of {fn.__name__} is variable"
        rv.append(param.name)
    return rv


def inject(**factories):
    params = {p for cm in factories.values() for p in get_params(cm)}

    def decorator(fn):
        fn_params = get_params(fn)

        @contextmanager
        def wrapper(**kwargs):
            for kw in kwargs:
                if kw not in params:
                    raise TypeError(
                        f"{fn.__name__}({', '.join(sorted(params))}) "
                        f"got an unexpected keyword argument {kw!r}"
                    )

            with ExitStack() as stack:
                fn_kwargs = {}
                for cm_name, cm in factories.items():
                    try:
                        cm_kwargs = {p: kwargs[p] for p in get_params(cm)}
                    except KeyError as e:
                        raise TypeError(
                            f"{fn.__name__}({', '.join(sorted(params))}) "
                            f"missing required argument {e} (from {cm.__name__})"
                        ) from None

                    cm_val = stack.enter_context(cm(**cm_kwargs))
                    if cm_name in fn_params:
                        fn_kwargs[cm_name] = cm_val

                yield partial(fn, **fn_kwargs)

        return wrapper

    return decorator


NUM_FEEDS = 8


def make_reader_with_entries(path, num_entries, num_feeds=NUM_FEEDS, text=False):
    reader = make_reader(path)
    reader._parser = parser = Parser()
    parser.tzinfo = None

    for i in range(num_feeds):
        feed = parser.feed(i, datetime(2010, 1, 1))
        reader.add_feed(feed.url)

    random.seed(0)
    for i in range(num_entries):
        kwargs = {}
        if text:
            kwargs.update(
                title=generate_lorem_ipsum(html=False, n=1, min=1, max=10),
                summary=generate_lorem_ipsum(html=False),
            )
        parser.entry(i % num_feeds, i, datetime(2010, 1, 1) + timedelta(i), **kwargs)

    return reader


def make_test_client(path):
    app = create_app(make_reader_config({'reader': {'url': path}}))
    client = app.test_client()
    with app.app_context():
        get_reader()
    return client


@contextmanager
def setup_db():
    with tempfile.TemporaryDirectory() as tmpdir:
        yield os.path.join(tmpdir, 'db.sqlite')


EXISTING_DB_PATH = None


@contextmanager
def setup_db_with_entries(num_entries):
    if EXISTING_DB_PATH:
        yield EXISTING_DB_PATH
        return
    with setup_db() as path:
        make_reader_with_entries(path, num_entries).update_feeds()
        yield path


@contextmanager
def setup_reader_with_entries(num_entries):
    with setup_db_with_entries(num_entries) as path:
        yield make_reader(path)


@contextmanager
def setup_client_with_entries(num_entries):
    with setup_db_with_entries(num_entries) as path:
        yield make_test_client(path)


@inject(reader=setup_reader_with_entries)
def time_get_entries_all(reader):
    for _ in reader.get_entries():
        pass


@inject(reader=setup_reader_with_entries)
def time_get_entries_read(reader):
    for _ in reader.get_entries(read=True):
        pass


@inject(reader=setup_reader_with_entries)
def time_get_entries_unread(reader):
    for _ in reader.get_entries(read=False):
        pass


@inject(reader=setup_reader_with_entries)
def time_get_entries_important(reader):
    for _ in reader.get_entries(important=True):
        pass


@inject(reader=setup_reader_with_entries)
def time_get_entries_unimportant(reader):
    for _ in reader.get_entries(important=False):
        pass


@inject(reader=setup_reader_with_entries)
def time_get_entries_enclosures(reader):
    for _ in reader.get_entries(has_enclosures=True):
        pass


@inject(reader=setup_reader_with_entries)
def time_get_entries_no_enclosures(reader):
    for _ in reader.get_entries(has_enclosures=False):
        pass


@inject(reader=setup_reader_with_entries)
def time_get_entries_feed(reader):
    feed = next(reader.get_feeds())
    for _ in reader.get_entries(feed=feed):
        pass


@inject(client=setup_client_with_entries)
def time_show(client):
    for _ in client.get('/?show=all').response:
        pass


@inject(client=setup_client_with_entries)
def time_show_100k(client):
    length = 0
    for chunk in client.get('/?show=all').response:
        length += len(chunk)
        if length >= 100000:
            break


@contextmanager
def setup_reader_with_fake_parser(num_entries):
    with setup_db() as path:
        yield make_reader_with_entries(path, num_entries)


@inject(reader=setup_reader_with_fake_parser)
def time_update_feeds(reader):
    reader.update_feeds()


@contextmanager
def setup_reader_feed_new(num_entries):
    with setup_db() as path:
        yield make_reader_with_entries(path, num_entries, num_feeds=1)


@contextmanager
def setup_reader_feed_old(num_entries):
    with setup_reader_feed_new(num_entries) as reader:
        reader.update_feeds()
        yield reader


def raise_too_many_variables(reader):
    original = getattr(reader._storage, '_get_entries_for_update_one_query', None)

    def wrapper(*args):
        original(*args)
        raise sqlite3.OperationalError("too many SQL variables")

    reader._storage._get_entries_for_update_one_query = wrapper


@contextmanager
def setup_reader_feed_new_fallback(num_entries):
    with setup_reader_feed_new(num_entries) as reader:
        raise_too_many_variables(reader)
        yield reader


@contextmanager
def setup_reader_feed_old_fallback(num_entries):
    with setup_reader_feed_old(num_entries) as reader:
        raise_too_many_variables(reader)
        yield reader


def _time_update_feed(reader):
    feed_url = list(reader._parser.feeds.values())[0].url
    reader.update_feed(feed_url)


time_update_feed_new = inject(reader=setup_reader_feed_new)(_time_update_feed)
time_update_feed_new_fallback = inject(reader=setup_reader_feed_new_fallback)(
    _time_update_feed
)
time_update_feed_old = inject(reader=setup_reader_feed_old)(_time_update_feed)
time_update_feed_old_fallback = inject(reader=setup_reader_feed_old_fallback)(
    _time_update_feed
)


@contextmanager
def setup_reader_with_text_entries(num_entries):
    if EXISTING_DB_PATH:
        yield make_reader(EXISTING_DB_PATH)
        return
    with setup_db() as path:
        reader = make_reader_with_entries(path, num_entries, text=True)
        reader.update_feeds()
        yield reader


@contextmanager
def setup_reader_with_search_and_some_read_entries(num_entries):
    if EXISTING_DB_PATH:
        yield make_reader(EXISTING_DB_PATH)
        return
    with setup_reader_with_text_entries(num_entries) as reader:
        reader.enable_search()
        reader.update_search()
        for i, entry in enumerate(reader.get_entries()):
            if i % 2 == 5:
                reader.mark_as_read(entry)
        yield reader


SEARCH_ENTRIES_QUERY = 'porta justo scelerisque dignissim convallis primis lacus'


@inject(reader=setup_reader_with_search_and_some_read_entries)
def time_search_entries_relevant_all(reader):
    for _ in reader.search_entries(SEARCH_ENTRIES_QUERY):
        pass


@inject(reader=setup_reader_with_search_and_some_read_entries)
def time_search_entries_relevant_read(reader):
    for _ in reader.search_entries(SEARCH_ENTRIES_QUERY, read=True):
        pass


@inject(reader=setup_reader_with_search_and_some_read_entries)
def time_search_entries_recent_all(reader):
    for _ in reader.search_entries(SEARCH_ENTRIES_QUERY, sort='recent'):
        pass


@inject(reader=setup_reader_with_search_and_some_read_entries)
def time_search_entries_recent_read(reader):
    for _ in reader.search_entries(SEARCH_ENTRIES_QUERY, sort='recent', read=True):
        pass


@inject(reader=setup_reader_with_text_entries)
def time_update_search(reader):
    # Sadly time() doesn't allow running the setup for every repeat,
    # so we enable/disable search inside the benchmark
    # (otherwise the second update_search() call has nothing to do).
    reader.enable_search()
    reader.update_search()
    reader.disable_search()


TIMINGS = OrderedDict(
    (tn.partition('_')[2], t)
    for tn, t in sorted(globals().items())
    if tn.startswith('time_')
)
TIMINGS_PARAMS_LIST = [(2 ** i,) for i in range(5, 12)]
TIMINGS_NUMBER = 4
PROFILE_PARAMS = TIMINGS_PARAMS_LIST[-1]
PARAM_IDS = ('num_entries',)


@click.group()
def cli():
    pass


@cli.command(name='list')
def list_():
    for timing in TIMINGS:
        print(timing)


def make_header(extra, names):
    return ' '.join(extra + names)


def make_row_fmt(extra, names, num_fmt='.3f'):
    extra_fmt = ['{{:>{}}}'.format(len(e)) for e in extra]
    names_fmt = ['{{:>{}{}}}'.format(len(n), num_fmt) for n in names]
    return ' '.join(extra_fmt + names_fmt)


@cli.command()
@click.argument('which', nargs=-1)
@click.option('-n', '--number', type=int, default=TIMINGS_NUMBER, show_default=True)
@click.option('-r', '--repeat', type=int, show_default=True)
@click.option(
    '--db',
    type=click.Path(exists=True, dir_okay=False),
    help="Use an existing database instead of generating ones of varying sizes. "
    "WARNING: Benchmarks that mutate the database might not make sense.",
)
@click.option('--query', help="Query for search_entries() timings; useful with --db.")
def time(which, number, repeat, db, query):
    if not which:
        which = ['*']

    if not repeat:
        extra = ['number'] + list(PARAM_IDS)
        timeit_func = timeit.timeit
        stats = {'': lambda x: x}
    else:
        extra = ['stat', 'number', 'repeat'] + list(PARAM_IDS)
        timeit_func = partial(timeit.repeat, repeat=repeat)

        # statistics.quantiles only gets added in Python 3.8
        import numpy as np

        stats = {
            'avg': np.mean,
            'min': lambda xs: min(xs),
            'p50': partial(np.quantile, q=0.5),
            'p90': partial(np.quantile, q=0.9),
        }

    names = [name for name in TIMINGS if any(fnmatchcase(name, w) for w in which)]

    header = make_header(extra, names)
    row_fmt = make_row_fmt(extra, names)

    if db:
        # HACK: use an existing DB
        # (for which num_entries can't vary, since it exists already)
        global EXISTING_DB_PATH
        EXISTING_DB_PATH = db
        global TIMINGS_PARAMS_LIST
        TIMINGS_PARAMS_LIST = [(0,)]
    if query:
        global SEARCH_ENTRIES_QUERY
        SEARCH_ENTRIES_QUERY = query

    def get_results():
        for params in TIMINGS_PARAMS_LIST:
            times = []
            for name in names:
                cm = TIMINGS[name](**dict(zip(PARAM_IDS, params)))
                with cm as fn:
                    time = timeit_func('fn()', globals=dict(fn=fn), number=number)
                times.append(time)
            yield list(params), times

    print(header)
    for params, results in get_results():
        for stat_name, stat in stats.items():
            if not repeat:
                prefix = [number]
            else:
                prefix = [stat_name, number, repeat]
            print(row_fmt.format(*prefix, *params, *map(stat, results)))


def fancy_division(a, b):
    try:
        return a / b
    except ZeroDivisionError:
        return math.copysign(float('inf'), a)


@cli.command()
@click.argument('before', type=click.File())
@click.argument('after', type=click.File())
@click.option(
    '--format',
    type=click.Choice(['percent-decrease', 'times']),
    default='percent-decrease',
    show_default=True,
    help="percent-decrease == 1 - after / before. times == after / before.",
)
def diff(before, after, format):
    if format == 'percent-decrease':
        row_num_fmt = '.1%'
        func = lambda b, a: 1 - fancy_division(float(a), float(b))
    elif format == 'times':
        # it would be nice if we could add an "x" after the number, but eh...
        row_num_fmt = '.2f'
        func = lambda b, a: fancy_division(float(a), float(b))
    else:
        assert False, "shouldn't happen"

    pairs = zip(before, after)

    b_line, a_line = next(pairs)
    assert b_line == a_line

    parts = b_line.split()
    first_param_index = parts.index(PARAM_IDS[0])
    first_name_index = first_param_index + len(PARAM_IDS)
    assert parts[first_param_index:first_name_index] == list(PARAM_IDS)

    extra = parts[:first_name_index]
    names = parts[first_name_index:]

    header = make_header(extra, names)
    row_fmt = make_row_fmt(extra, names, row_num_fmt)

    print(header)
    for b_line, a_line in pairs:
        b_parts, a_parts = b_line.split(), a_line.split()
        assert b_parts[:first_name_index] == a_parts[:first_name_index]

        results = [
            func(b, a)
            for a, b in zip(a_parts[first_name_index:], b_parts[first_name_index:])
        ]
        print(row_fmt.format(*b_parts[:first_name_index], *results))


@cli.command()
@click.argument('which', nargs=-1)
def profile(which):
    names = [name for name in TIMINGS if any(fnmatchcase(name, w) for w in which)]
    params = PROFILE_PARAMS

    for name in names:
        print(name, ' '.join('{}={}'.format(i, p) for i, p in zip(PARAM_IDS, params)))
        print()

        cm = TIMINGS[name](**dict(zip(PARAM_IDS, params)))

        pr = cProfile.Profile()
        with cm as fn:
            pr.enable()
            fn()
            pr.disable()
        pstats.Stats(pr).strip_dirs().sort_stats('cumulative').print_stats(40)


if __name__ == '__main__':
    cli()



import datetime

from reader import Content
from reader import Enclosure
from reader._types import EntryData
from reader._types import FeedData


feed = FeedData(
    url='{}relative.rss'.format(url_base), link='{}file.html'.format(rel_base)
)

entries = [
    EntryData(
        feed_url=feed.url,
        id='7bd204c6-1655-4c27-aeee-53f933c5395f',
        updated=None,
        link='{}blog/post/1'.format(rel_base),
        summary='one <a href="{}target">two</a> three'.format(rel_base),
        content=(
            Content(
                value='<script>evil</script> content', type='text/plain', language=None
            ),
            Content(value='content', type='text/html', language=None),
        ),
        enclosures=(
            # for RSS feedparser doesn't make relative links absolute
            # (it does for Atom)
            Enclosure(href='enclosure?q=a#fragment'),
        ),
    )
]



import builtins
import itertools
import logging
import numbers
import warnings
from contextlib import nullcontext
from datetime import datetime
from types import MappingProxyType
from typing import Any
from typing import Callable
from typing import Collection
from typing import Iterable
from typing import Iterator
from typing import Mapping
from typing import MutableSequence
from typing import Optional
from typing import overload
from typing import Tuple
from typing import TypeVar
from typing import Union

import reader._updater
from ._parser import default_parser
from ._parser import Parser
from ._parser import SESSION_TIMEOUT
from ._requests_utils import TimeoutType
from ._search import Search
from ._storage import Storage
from ._types import DEFAULT_RESERVED_NAME_SCHEME
from ._types import EntryData
from ._types import EntryFilterOptions
from ._types import EntryUpdateIntent
from ._types import FeedFilterOptions
from ._types import FeedForUpdate
from ._types import FeedUpdateIntent
from ._types import fix_datetime_tzinfo
from ._types import NameScheme
from ._types import ParsedFeed
from ._utils import make_pool_map
from ._utils import zero_or_one
from .exceptions import EntryNotFoundError
from .exceptions import FeedMetadataNotFoundError
from .exceptions import FeedNotFoundError
from .exceptions import InvalidPluginError
from .exceptions import ParseError
from .plugins import _DEFAULT_PLUGINS
from .plugins import _PLUGINS
from .types import _entry_argument
from .types import _feed_argument
from .types import Entry
from .types import EntryCounts
from .types import EntryInput
from .types import EntrySearchCounts
from .types import EntrySearchResult
from .types import EntrySortOrder
from .types import EntryUpdateStatus
from .types import Feed
from .types import FeedCounts
from .types import FeedInput
from .types import FeedSortOrder
from .types import JSONType
from .types import MISSING
from .types import MissingType
from .types import SearchSortOrder
from .types import TagFilterInput
from .types import UpdatedFeed
from .types import UpdateResult


log = logging.getLogger('reader')


_T = TypeVar('_T')
_U = TypeVar('_U')

ReaderPluginType = Callable[['Reader'], None]
AfterEntryUpdateHook = Callable[
    ['Reader', EntryData[datetime], EntryUpdateStatus], None
]
_PostFeedUpdatePluginType = Callable[['Reader', str], None]


def make_reader(
    url: str,
    *,
    feed_root: Optional[str] = None,
    plugins: Iterable[Union[str, ReaderPluginType]] = _DEFAULT_PLUGINS,
    session_timeout: TimeoutType = SESSION_TIMEOUT,
    reserved_name_scheme: Mapping[str, str] = DEFAULT_RESERVED_NAME_SCHEME,
    _storage: Optional[Storage] = None,
    _storage_factory: Any = None,
) -> 'Reader':
    """Create a new :class:`Reader`.

    *reader* can optionally parse local files, with the feed URL either
    a bare path or a file URI.

    The interpretation of local feed URLs depends on the value of the
    feed ``feed_root`` argument.
    It can be one of the following:

    ``None``

        No local file parsing. Updating local feeds will fail.

    ``''`` (the empty string)

        Full filesystem access. This should be used only if the source of
        feed URLs is trusted.

        Both absolute and relative feed paths are supported.
        The current working directory is used normally
        (as if the path was passed to :func:`open`).

        Example: Assuming the current working directory is ``/feeds``,
        all of the following feed URLs correspond to ``/feeds/feed.xml``:
        ``feed.xml``, ``/feeds/feed.xml``, ``file:feed.xml``,
        and ``file:/feeds/feed.xml``.

    ``'/path/to/feed/root'`` (any non-empty string)

        An absolute path; all feed URLs are interpreted as relative to it.
        This can be used if the source  of feed URLs is untrusted.

        Feed paths must be relative. The current working directory is ignored.

        Example: Assuming the feed root is ``/feeds``, feed URLs
        ``feed.xml`` and ``file:feed.xml`` correspond to ``/feeds/feed.xml``.
        ``/feed.xml`` and ``file:/feed.xml`` are both errors.

        Relative paths pointing outside the feed root are errors,
        to prevent directory traversal attacks. Note that symbolic links
        inside the feed root *can* point outside it.

        The root and feed paths are joined and normalized with no regard for
        symbolic links; see :func:`os.path.normpath` for details.

        Accessing device files on Windows is an error.

    Args:
        url (str): Path to the reader database.

        feed_root (str or None):
            Directory where to look for local feeds.
            One of ``None`` (don't open local feeds; default),
            ``''`` (full filesystem access), or
            ``'/path/to/feed/root'``
            (an absolute path that feed paths are relative to).

        plugins (iterable(str or callable(Reader)) or None):
            An iterable of built-in plugin names or
            `plugin(reader) --> None` callables.
            The callables are called with the reader object
            before it is returned.
            Exceptions from plugin code will propagate to the caller.
            The only plugin used by default is ``reader.ua_fallback``.

        session_timeout (float or tuple(float, float) or None):
            When retrieving HTTP(S) feeds,
            how many seconds to wait for the server to send data,
            as a float, or a (connect timeout, read timeout) tuple.
            Passed to the underlying `Requests session`_.

        reserved_name_scheme (dict(str, str) or None):
            Value for :attr:`~Reader.reserved_name_scheme`.
            The prefixes default to ``.reader.``/``.plugin.``,
            and the separator to ``.``

    .. _Requests session: https://requests.readthedocs.io/en/master/user/advanced/#timeouts

    Returns:
        Reader: The reader.

    Raises:
        StorageError
        InvalidPluginError: An invalid plugin name was passed to ``plugins``.
        ReaderError: An ambiguous exception ocurred while creating the reader.

    .. versionadded:: 1.6
        The ``feed_root`` keyword argument.

    .. versionadded:: 1.14
        The ``session_timeout`` keyword argument,
        with a default of (3.05, 60) seconds;
        the previous behavior was to *never time out*.

    .. versionadded:: 1.16
        The ``plugins`` keyword argument. Using an invalid plugin name
        raises :exc:`InvalidPluginError`, a :exc:`ValueError` subclass.

    .. versionadded:: 1.17
        The ``reserved_name_scheme`` argument.

    .. versionchanged:: 2.0
        ``feed_root`` now defaults to ``None`` (don't open local feeds)
        instead of ``''`` (full filesystem access).

    """

    # If we ever need to change the signature of make_reader(),
    # or support additional storage/search implementations,
    # we'll need to do the wiring differently.
    #
    # See this comment for details on how it should evolve:
    # https://github.com/lemon24/reader/issues/168#issuecomment-642002049

    storage = _storage or Storage(url, factory=_storage_factory)

    # For now, we're using a storage-bound search provider.
    search = Search(storage)

    parser = default_parser(feed_root, session_timeout=session_timeout)

    reader = Reader(
        storage, search, parser, reserved_name_scheme, _called_directly=False
    )

    for plugin in plugins:
        if isinstance(plugin, str):
            if plugin not in _PLUGINS:
                raise InvalidPluginError(f"no such built-in plugin: {plugin!r}")

            plugin_func = _PLUGINS[plugin]
        else:
            plugin_func = plugin

        try:
            plugin_func(reader)  # type: ignore
        except Exception:  # pragma: no cover
            # TODO: this whole branch is not tested
            reader.close()
            # TODO: this should raise a custom exception (but can't because of backwards compatibility)
            raise

    return reader


class Reader:

    """A feed reader.

    Persists feed and entry state, provides operations on them,
    and stores configuration.

    Currently, the following feed types are supported:

    * Atom (provided by `feedparser`_)
    * RSS (provided by `feedparser`_)
    * JSON Feed

    .. _feedparser: https://feedparser.readthedocs.io/en/latest/


    .. important::

        Reader objects should be created using :func:`make_reader`; the Reader
        constructor is not stable yet and may change without any notice.

    .. important::

        The :class:`Reader` object is not thread safe;
        its methods should be called only from the thread that created it.

        To access the same database from multiple threads,
        create one instance in each thread.
        If you have a strong use case preventing you to do so,
        please +1 / comment in :issue:`206`.


    .. versionadded:: 1.13
        JSON Feed support.


    """

    def __init__(
        self,
        _storage: Storage,
        _search: Search,
        _parser: Parser,
        _reserved_name_scheme: Mapping[str, str],
        _called_directly: bool = True,
    ):
        self._storage = _storage
        self._search = _search
        self._parser = _parser

        try:
            self.reserved_name_scheme = _reserved_name_scheme
        except AttributeError as e:
            raise ValueError(str(e)) from (e.__cause__ or e)

        self._updater = reader._updater
        self._post_feed_update_plugins: Collection[_PostFeedUpdatePluginType] = []

        #: List of functions called for each updated entry
        #: after the feed was updated.
        #:
        #: Each function is called with:
        #:
        #: * `reader` – the :class:`Reader` instance
        #: * `entry` – an :class:`Entry`-like object
        #: * `status` – an :class:`EntryUpdateStatus` value
        #:
        #: Each function should return :const:`None`.
        #:
        #: .. warning::
        #:
        #:  The only `entry` attributes guaranteed to be present are
        #:  :attr:`~Entry.feed_url`, :attr:`~Entry.id`,
        #:  and :attr:`~Entry.object_id`;
        #:  all other attributes may be missing
        #:  (accessing them may raise :exc:`AttributeError`).
        #:
        #: .. versionadded:: 1.20
        #:
        self.after_entry_update_hooks: MutableSequence[AfterEntryUpdateHook] = []

        if _called_directly:
            warnings.warn(
                "Reader objects should be created using make_reader(); the Reader "
                "constructor is not stable yet and may change without any notice.",
            )

    def close(self) -> None:
        """Close this :class:`Reader`.

        Releases any underlying resources associated with the reader.

        The reader becomes unusable from this point forward;
        a :exc:`ReaderError` will be raised if any other method is called.

        Raises:
            ReaderError

        """
        self._storage.close()

    def add_feed(self, feed: FeedInput) -> None:
        """Add a new feed.

        Feed updates are enabled by default.

        Args:
            feed (str or Feed): The feed URL.

        Raises:
            FeedExistsError
            StorageError

        """
        url = _feed_argument(feed)
        now = self._now()
        self._storage.add_feed(url, now)

    def delete_feed(self, feed: FeedInput) -> None:
        """Delete a feed and all of its entries, metadata, and tags.

        Args:
            feed (str or Feed): The feed URL.

        Raises:
            FeedNotFoundError
            StorageError

        .. versionadded:: 1.18
            Renamed from :meth:`remove_feed`.

        """
        url = _feed_argument(feed)
        self._storage.delete_feed(url)

    def change_feed_url(self, old: FeedInput, new: FeedInput) -> None:
        """Change the URL of a feed.

        User-defined feed attributes are preserved:
        :attr:`~Feed.added`, :attr:`~Feed.user_title`.
        Feed-defined feed attributes are also preserved,
        at least until the next update:
        :attr:`~Feed.title`, :attr:`~Feed.link`, :attr:`~Feed.author`
        (except :attr:`~Feed.updated`, which gets set to None).
        All other feed attributes are set to their default values.

        The entries, tags and metadata are preserved.

        Args:
            old (str or Feed): The old feed; must exist.
            new (str or Feed): The new feed; must not exist.

        Raises:
            FeedNotFoundError: If ``old`` does not exist.
            FeedExistsError: If ``new`` already exists.
            StorageError

        .. versionadded:: 1.8

        """
        self._storage.change_feed_url(_feed_argument(old), _feed_argument(new))

    def get_feeds(
        self,
        *,
        feed: Optional[FeedInput] = None,
        tags: TagFilterInput = None,
        broken: Optional[bool] = None,
        updates_enabled: Optional[bool] = None,
        sort: FeedSortOrder = 'title',
        limit: Optional[int] = None,
        starting_after: Optional[FeedInput] = None,
    ) -> Iterable[Feed]:
        """Get all or some of the feeds.

        The ``tags`` argument can be a list of one or more feed tags.
        Multiple tags are interpreted as a conjunction (AND).
        To use a disjunction (OR), use a nested list.
        To negate a tag, prefix the tag value with a minus sign (``-``).
        Examples:

        ``['one']``

            one

        ``['one', 'two']``
        ``[['one'], ['two']]``

            one AND two

        ``[['one', 'two']]``

            one OR two

        ``[['one', 'two'], 'three']``

            (one OR two) AND three

        ``['one', '-two']``

            one AND NOT two

        Special values :const:`True` and :const:`False`
        match feeds with any tags and no tags, respectively.

        ``True``
        ``[True]``

            *any tags*

        ``False``
        ``[False]``

            *no tags*

        ``[True, '-one']``

            *any tags* AND NOT one

        ``[[False, 'one']]``

            *no tags* OR one

        Args:
            feed (str or Feed or None): Only return the feed with this URL.
            tags (None or bool or list(str or bool or list(str or bool))):
                Only return feeds matching these tags.
            broken (bool or None): Only return broken / healthy feeds.
            updates_enabled (bool or None):
                Only return feeds that have updates enabled / disabled.
            sort (str): How to order feeds; one of ``'title'`` (by
                :attr:`~Feed.user_title` or :attr:`~Feed.title`, case
                insensitive; default), or ``'added'`` (last added first).
            limit (int or None): A limit on the number of feeds to be returned;
                by default, all feeds are returned.
            starting_after (str or Feed or None):
                Return feeds after this feed; a cursor for use in pagination.

        Yields:
            :class:`Feed`: Sorted according to ``sort``.

        Raises:
            StorageError
            FeedNotFoundError: If ``starting_after`` does not exist.

        .. versionadded:: 1.7
            The ``tags`` keyword argument.

        .. versionadded:: 1.7
            The ``broken`` keyword argument.

        .. versionadded:: 1.11
            The ``updates_enabled`` keyword argument.

        .. versionadded:: 1.12
            The ``limit`` and ``starting_after`` keyword arguments.

        """
        filter_options = FeedFilterOptions.from_args(
            feed, tags, broken, updates_enabled
        )

        if sort not in ('title', 'added'):
            raise ValueError("sort should be one of ('title', 'added')")

        if limit is not None:
            if not isinstance(limit, numbers.Integral) or limit < 1:
                raise ValueError("limit should be a positive integer")

        rv = self._storage.get_feeds(
            filter_options,
            sort,
            limit,
            _feed_argument(starting_after) if starting_after else None,
        )

        for rv_feed in rv:
            yield fix_datetime_tzinfo(rv_feed, 'updated', 'added', 'last_updated')

    @overload
    def get_feed(self, feed: FeedInput) -> Feed:  # pragma: no cover
        ...

    @overload
    def get_feed(
        self, feed: FeedInput, default: _T
    ) -> Union[Feed, _T]:  # pragma: no cover
        ...

    def get_feed(
        self, feed: FeedInput, default: Union[MissingType, _T] = MISSING
    ) -> Union[Feed, _T]:
        """Get a feed.

        Like ``next(iter(reader.get_feeds(feed=feed)), default)``,
        but raises a custom exception instead of :exc:`StopIteration`.

        Arguments:
            feed (str or Feed): The feed URL.
            default: Returned if given and the feed does not exist.

        Returns:
            Feed: The feed.

        Raises:
            FeedNotFoundError
            StorageError

        """
        return zero_or_one(
            self.get_feeds(feed=feed),
            lambda: FeedNotFoundError(_feed_argument(feed)),
            default,
        )

    def get_feed_counts(
        self,
        *,
        feed: Optional[FeedInput] = None,
        tags: TagFilterInput = None,
        broken: Optional[bool] = None,
        updates_enabled: Optional[bool] = None,
    ) -> FeedCounts:
        """Count all or some of the feeds.

        See :meth:`~Reader.get_feeds()` for details on how filtering works.

        Args:
            feed (str or Feed or None): Only count the feed with this URL.
            tags (None or bool or list(str or bool or list(str or bool))):
                Only count feeds matching these tags.
            broken (bool or None): Only count broken / healthy feeds.
            updates_enabled (bool or None):
                Only count feeds that have updates enabled / disabled.

        Returns:
            FeedCounts:

        Raises:
            StorageError

        .. versionadded:: 1.11

        """
        filter_options = FeedFilterOptions.from_args(
            feed, tags, broken, updates_enabled
        )
        return self._storage.get_feed_counts(filter_options)

    def set_feed_user_title(self, feed: FeedInput, title: Optional[str]) -> None:
        """Set a user-defined title for a feed.

        Args:
            feed (str or Feed): The feed URL.
            title (str or None): The title, or None to remove the current title.

        Raises:
            FeedNotFoundError
            StorageError

        """
        url = _feed_argument(feed)
        return self._storage.set_feed_user_title(url, title)

    def enable_feed_updates(self, feed: FeedInput) -> None:
        """Enable updates for a feed.

        See :meth:`~Reader.update_feeds` for details.

        Args:
            feed (str or Feed): The feed URL.

        Raises:
            FeedNotFoundError
            StorageError

        .. versionadded:: 1.11

        """
        url = _feed_argument(feed)
        self._storage.set_feed_updates_enabled(url, True)

    def disable_feed_updates(self, feed: FeedInput) -> None:
        """Disable updates for a feed.

        See :meth:`~Reader.update_feeds` for details.

        Args:
            feed (str or Feed): The feed URL.

        Raises:
            FeedNotFoundError
            StorageError

        .. versionadded:: 1.11

        """
        url = _feed_argument(feed)
        self._storage.set_feed_updates_enabled(url, False)

    def update_feeds(
        self,
        *,
        new: Optional[bool] = None,
        workers: int = 1,
    ) -> None:
        """Update all the feeds that have updates enabled.

        Silently skip feeds that raise :exc:`ParseError`.

        Roughly equivalent to ``for _ in reader.update_feed_iter(...): pass``.

        Args:
            new (bool or None):
                Only update feeds that have never been updated
                / have been updated before. Defaults to None.
            workers (int): Number of threads to use when getting the feeds.

        Raises:
            StorageError

        .. versionchanged:: 1.11
            Only update the feeds that have updates enabled.

        .. versionchanged:: 1.15
            Update entries whenever their content changes,
            regardless of their :attr:`~Entry.updated` date.

            Content-only updates (not due to an :attr:`~Entry.updated` change)
            are limited to 24 consecutive updates,
            to prevent spurious updates for entries whose content changes
            excessively (for example, because it includes the current time).

            Previously, entries would be updated only if the
            entry :attr:`~Entry.updated` was *newer* than the stored one.

        .. versionchanged:: 2.0
            Removed the ``new_only`` parameter.

        .. versionchanged:: 2.0
            All parameters are keyword-only.

        """
        for url, value in self.update_feeds_iter(new=new, workers=workers):
            if isinstance(value, ParseError):
                log.exception(
                    "update feed %r: error while getting/parsing feed, "
                    "skipping; exception: %r",
                    url,
                    value.__cause__,
                    exc_info=value,
                )
                continue

            assert not isinstance(value, Exception), value

    def update_feeds_iter(
        self,
        *,
        new: Optional[bool] = None,
        workers: int = 1,
    ) -> Iterable[UpdateResult]:
        """Update all the feeds that have updates enabled.

        Args:
            new (bool or None):
                Only update feeds that have never been updated
                / have been updated before. Defaults to None.
            workers (int): Number of threads to use when getting the feeds.


        Yields:
            :class:`UpdateResult`:
                An (url, value) pair; the value is one of:

                * a summary of the updated feed, if the update was successful
                * None, if the server indicated the feed has not changed
                  since the last update
                * an exception instance

                Currently, the exception is always a :exc:`ParseError`,
                but other :exc:`ReaderError` subclasses may be yielded
                in the future.

        Raises:
            StorageError

        .. versionadded:: 1.14

        .. versionchanged:: 1.15
            Update entries whenever their content changes.
            See :meth:`~Reader.update_feeds` for details.

        .. versionchanged:: 2.0
            Removed the ``new_only`` parameter.

        .. versionchanged:: 2.0
            All parameters are keyword-only.

        """
        if workers < 1:
            raise ValueError("workers must be a positive integer")

        make_map = nullcontext(builtins.map) if workers == 1 else make_pool_map(workers)

        with make_map as map:
            results = self._update_feeds(new=new, map=map)

            for url, value in results:
                if isinstance(value, FeedNotFoundError):
                    log.info("update feed %r: feed removed during update", url)
                    continue

                if isinstance(value, Exception):
                    if not isinstance(value, ParseError):
                        raise value

                yield UpdateResult(url, value)

    def update_feed(self, feed: FeedInput) -> Optional[UpdatedFeed]:
        """Update a single feed.

        The feed will be updated even if updates are disabled for it.

        Args:
            feed (str or Feed): The feed URL.

        Returns:
            UpdatedFeed or None:
            A summary of the updated feed or None,
            if the server indicated the feed has not changed
            since the last update.

        Raises:
            FeedNotFoundError
            ParseError
            StorageError

        .. versionchanged:: 1.14
            The method now returns UpdatedFeed or None instead of None.

        .. versionchanged:: 1.15
            Update entries whenever their content changes.
            See :meth:`~Reader.update_feeds` for details.

        """
        url = _feed_argument(feed)
        _, rv = zero_or_one(
            self._update_feeds(url=url, enabled_only=False),
            lambda: FeedNotFoundError(url),
        )
        if isinstance(rv, Exception):
            raise rv
        return rv

    @staticmethod
    def _now() -> datetime:
        return datetime.utcnow()

    # The type of map should be
    #
    #   Callable[[Callable[[_T], _U], Iterable[_T]], Iterator[_U]]
    #
    # but mypy gets confused; known issue:
    #
    # https://github.com/python/mypy/issues/1317
    # https://github.com/python/mypy/issues/6697

    def _update_feeds(
        self,
        url: Optional[str] = None,
        new: Optional[bool] = None,
        enabled_only: bool = True,
        map: Callable[[Callable[[Any], Any], Iterable[Any]], Iterator[Any]] = map,
    ) -> Iterator[Tuple[str, Union[UpdatedFeed, None, Exception]]]:

        # global_now is used as first_updated_epoch for all new entries,
        # so that the subset of new entries from an update appears before
        # all others and the entries in it are sorted by published/updated;
        # if we used last_updated (now) for this, they would be sorted
        # by feed order first (due to now increasing for each feed).
        #
        # A side effect of relying first_updated_epoch for ordering is that
        # for the second of two new feeds updated in the same update_feeds()
        # call, first_updated_epoch != last_updated.
        #
        global_now = self._now()

        # Excluding the special exception handling,
        # this function is a pipeline that looks somewhat like this:
        #
        #   self._storage.get_feeds_for_update \
        #   | self._updater.process_old_feed \
        #   | xargs -n1 -P $workers self._parse_feed_for_update \
        #   | self._get_entries_for_update \
        #   | self._updater.make_update_intents \
        #   | self._update_feed
        #
        # Since we only need _parse_feed_for_update to run in parallel,
        # everything after that is in a single for loop for readability.
        #
        # It may make sense to also have _get_entries_for_update run in
        # parallel with a different (slower) storage, but for now we're good.

        feeds_for_update = self._storage.get_feeds_for_update(url, new, enabled_only)
        feeds_for_update = builtins.map(
            self._updater.process_old_feed, feeds_for_update
        )

        pairs = map(self._parse_feed_for_update, feeds_for_update)

        for feed_for_update, parse_result in pairs:
            rv: Union[UpdatedFeed, None, Exception]

            try:
                # give storage a chance to consume the entries in a streaming fashion;
                parsed_entries = itertools.tee(
                    parse_result.entries
                    if parse_result and not isinstance(parse_result, Exception)
                    else ()
                )
                entry_pairs = zip(
                    parsed_entries[0],
                    self._storage.get_entries_for_update(
                        (e.feed_url, e.id) for e in parsed_entries[1]
                    ),
                )

                now = self._now()
                (
                    feed_to_update,
                    entries_to_update,
                ) = self._updater.make_update_intents(
                    feed_for_update, now, global_now, parse_result, entry_pairs
                )

                counts = self._update_feed(
                    feed_for_update.url, feed_to_update, entries_to_update
                )

                if isinstance(parse_result, Exception):
                    rv = parse_result
                elif parse_result:
                    rv = UpdatedFeed(feed_for_update.url, *counts)
                else:
                    rv = None

            except Exception as e:
                rv = e

            yield feed_for_update.url, rv

    def _parse_feed_for_update(
        self, feed: FeedForUpdate
    ) -> Tuple[FeedForUpdate, Union[ParsedFeed, None, ParseError]]:
        try:
            return feed, self._parser(feed.url, feed.http_etag, feed.http_last_modified)
        except ParseError as e:
            log.debug(
                "_parse_feed_for_update exception, traceback follows", exc_info=True
            )
            return feed, e

    def _update_feed(
        self,
        url: str,
        feed_to_update: Optional[FeedUpdateIntent],
        entries_to_update: Iterable[EntryUpdateIntent],
    ) -> Tuple[int, int]:
        if feed_to_update:
            if entries_to_update:
                self._storage.add_or_update_entries(entries_to_update)
            self._storage.update_feed(feed_to_update)

        # if feed_for_update.url != parsed_feed.feed.url, the feed was redirected.
        # TODO: Maybe handle redirects somehow else (e.g. change URL if permanent).

        for feed_plugin in self._post_feed_update_plugins:
            feed_plugin(self, url)

        new_count = 0
        updated_count = 0
        for entry in entries_to_update:
            if entry.new:
                new_count += 1
                entry_status = EntryUpdateStatus.NEW
            else:
                updated_count += 1
                entry_status = EntryUpdateStatus.MODIFIED
            for entry_hook in self.after_entry_update_hooks:
                entry_hook(self, entry.entry, entry_status)

        return new_count, updated_count

    def get_entries(
        self,
        *,
        feed: Optional[FeedInput] = None,
        entry: Optional[EntryInput] = None,
        read: Optional[bool] = None,
        important: Optional[bool] = None,
        has_enclosures: Optional[bool] = None,
        feed_tags: TagFilterInput = None,
        sort: EntrySortOrder = 'recent',
        limit: Optional[int] = None,
        starting_after: Optional[EntryInput] = None,
    ) -> Iterable[Entry]:
        """Get all or some of the entries.

        Entries are sorted according to ``sort``. Possible values:

        ``'recent'``

            Most recent first. Currently, that means:

            * by import date for entries published less than 7 days ago
            * by published date otherwise (if an entry does not have
              :attr:`~Entry.published`, :attr:`~Entry.updated` is used)

            This is to make sure newly imported entries appear at the top
            regardless of when the feed says they were published
            (sometimes, it lies by a day or two).

            .. note::

                The algorithm for "recent" is a heuristic and may change over time.

        ``'random'``

            Random order (shuffled). At at most 256 entries will be returned.

            .. versionadded:: 1.2

        Args:
            feed (str or Feed or None): Only return the entries for this feed.
            entry (tuple(str, str) or Entry or None):
                Only return the entry with this (feed URL, entry id) tuple.
            read (bool or None): Only return (un)read entries.
            important (bool or None): Only return (un)important entries.
            has_enclosures (bool or None): Only return entries that (don't)
                have enclosures.
            feed_tags (None or bool or list(str or bool or list(str or bool))):
                Only return the entries from feeds matching these tags;
                works like the :meth:`~Reader.get_feeds()` ``tags`` argument.
            sort (str): How to order entries; one of ``'recent'`` (default)
                or ``'random'``.
            limit (int or None): A limit on the number of entries to be returned;
                by default, all entries are returned.
            starting_after (tuple(str, str) or Entry or None):
                Return entries after this entry; a cursor for use in pagination.
                Using ``starting_after`` with ``sort='random'`` is not supported.

        Yields:
            :class:`Entry`: Sorted according to ``sort``.

        Raises:
            StorageError
            EntryNotFoundError: If ``starting_after`` does not exist.

        .. versionadded:: 1.2
            The ``sort`` keyword argument.

        .. versionadded:: 1.7
            The ``feed_tags`` keyword argument.

        .. versionadded:: 1.12
            The ``limit`` and ``starting_after`` keyword arguments.

        """

        # If we ever implement pagination, consider following the guidance in
        # https://specs.openstack.org/openstack/api-wg/guidelines/pagination_filter_sort.html

        filter_options = EntryFilterOptions.from_args(
            feed, entry, read, important, has_enclosures, feed_tags
        )

        if sort not in ('recent', 'random'):
            raise ValueError("sort should be one of ('recent', 'random')")

        if limit is not None:
            if not isinstance(limit, numbers.Integral) or limit < 1:
                raise ValueError("limit should be a positive integer")

        if starting_after and sort == 'random':
            raise ValueError("using starting_after with sort='random' not supported")

        now = self._now()
        rv = self._storage.get_entries(
            now,
            filter_options,
            sort,
            limit,
            _entry_argument(starting_after) if starting_after else None,
        )

        for rv_entry in rv:
            yield fix_datetime_tzinfo(
                rv_entry,
                'updated',
                'published',
                'last_updated',
                feed=fix_datetime_tzinfo(
                    rv_entry.feed, 'updated', 'added', 'last_updated'
                ),
            )

    @overload
    def get_entry(self, entry: EntryInput) -> Entry:  # pragma: no cover
        ...

    @overload
    def get_entry(
        self, entry: EntryInput, default: _T
    ) -> Union[Entry, _T]:  # pragma: no cover
        ...

    def get_entry(
        self, entry: EntryInput, default: Union[MissingType, _T] = MISSING
    ) -> Union[Entry, _T]:
        """Get an entry.

        Like ``next(iter(reader.get_entries(entry=entry)), default)``,
        but raises a custom exception instead of :exc:`StopIteration`.

        Args:
            entry (tuple(str, str) or Entry): (feed URL, entry id) tuple.
            default: Returned if given and the entry does not exist.

        Returns:
            Entry: The entry.

        Raises:
            EntryNotFoundError
            StorageError

        """
        return zero_or_one(
            self.get_entries(entry=entry),
            lambda: EntryNotFoundError(*_entry_argument(entry)),
            default,
        )

    def get_entry_counts(
        self,
        *,
        feed: Optional[FeedInput] = None,
        entry: Optional[EntryInput] = None,
        read: Optional[bool] = None,
        important: Optional[bool] = None,
        has_enclosures: Optional[bool] = None,
        feed_tags: TagFilterInput = None,
    ) -> EntryCounts:
        """Count all or some of the entries.

        See :meth:`~Reader.get_entries()` for details on how filtering works.

        Args:
            feed (str or Feed or None): Only count the entries for this feed.
            entry (tuple(str, str) or Entry or None):
                Only count the entry with this (feed URL, entry id) tuple.
            read (bool or None): Only count (un)read entries.
            important (bool or None): Only count (un)important entries.
            has_enclosures (bool or None): Only count entries that (don't)
                have enclosures.
            feed_tags (None or bool or list(str or bool or list(str or bool))):
                Only count the entries from feeds matching these tags.

        Returns:
            EntryCounts:

        Raises:
            StorageError

        .. versionadded:: 1.11

        """

        filter_options = EntryFilterOptions.from_args(
            feed, entry, read, important, has_enclosures, feed_tags
        )
        now = self._now()
        return self._storage.get_entry_counts(now, filter_options)

    def mark_entry_as_read(self, entry: EntryInput) -> None:
        """Mark an entry as read.

        Args:
            entry (tuple(str, str) or Entry): (feed URL, entry id) tuple.

        Raises:
            EntryNotFoundError
            StorageError

        .. versionadded:: 1.18
            Renamed from :meth:`mark_as_read`.

        """
        feed_url, entry_id = _entry_argument(entry)
        self._storage.mark_as_read_unread(feed_url, entry_id, True)

    def mark_entry_as_unread(self, entry: EntryInput) -> None:
        """Mark an entry as unread.

        Args:
            entry (tuple(str, str) or Entry): (feed URL, entry id) tuple.

        Raises:
            EntryNotFoundError
            StorageError

        .. versionadded:: 1.18
            Renamed from :meth:`mark_as_unread`.

        """
        feed_url, entry_id = _entry_argument(entry)
        self._storage.mark_as_read_unread(feed_url, entry_id, False)

    def mark_entry_as_important(self, entry: EntryInput) -> None:
        """Mark an entry as important.

        Args:
            entry (tuple(str, str) or Entry): (feed URL, entry id) tuple.

        Raises:
            EntryNotFoundError
            StorageError

        .. versionadded:: 1.18
            Renamed from :meth:`mark_as_important`.

        """
        feed_url, entry_id = _entry_argument(entry)
        self._storage.mark_as_important_unimportant(feed_url, entry_id, True)

    def mark_entry_as_unimportant(self, entry: EntryInput) -> None:
        """Mark an entry as unimportant.

        Args:
            entry (tuple(str, str) or Entry): (feed URL, entry id) tuple.

        Raises:
            EntryNotFoundError
            StorageError

        .. versionadded:: 1.18
            Renamed from :meth:`mark_as_unimportant`.

        """
        feed_url, entry_id = _entry_argument(entry)
        self._storage.mark_as_important_unimportant(feed_url, entry_id, False)

    def get_feed_metadata(
        self,
        feed: FeedInput,
        key: Optional[str] = None,
    ) -> Iterable[Tuple[str, JSONType]]:
        """Get all or some of the metadata for a feed as ``(key, value)`` pairs.

        Args:
            feed (str or Feed): The feed URL.
            key (str or None): Only return the metadata for this key.

        Yields:
            tuple(str, JSONType): ``(key, value)`` pairs, in undefined order.
            JSONType is whatever :py:func:`json.dumps` accepts.

        Raises:
            StorageError

        .. versionchanged:: 1.18
            :meth:`get_feed_metadata` was renamed to :meth:`get_feed_metadata_item`,
            :meth:`iter_feed_metadata` was renamed to :meth:`get_feed_metadata`.

        .. versionchanged:: 2.0
            The ``get_feed_metadata(feed, key, default=no value, /)``
            (positional arguments only)
            :meth:`get_feed_metadata_item` alias was removed.

        """

        # get_feed_metadata(feed, *, key=None) -> (key, value), ...
        feed_url = _feed_argument(feed)
        return self._storage.iter_metadata((feed_url,), key)

    @overload
    def get_feed_metadata_item(
        self, feed: FeedInput, key: str
    ) -> JSONType:  # pragma: no cover
        ...

    @overload
    def get_feed_metadata_item(
        self, feed: FeedInput, key: str, default: _T
    ) -> Union[JSONType, _T]:  # pragma: no cover
        ...

    def get_feed_metadata_item(
        self, feed: FeedInput, key: str, default: Union[MissingType, _T] = MISSING
    ) -> Union[JSONType, _T]:
        """Get metadata for a feed.

        Like ``next(iter(reader.get_feed_metadata(feed, key=key)), (None, default))[1]``,
        but raises a custom exception instead of :exc:`StopIteration`.

        Args:
            feed (str or Feed): The feed URL.
            key (str): The key of the metadata to retrieve.
            default: Returned if given and no metadata exists for `key`.

        Returns:
            JSONType: The metadata value.
            JSONType is whatever :py:func:`json.dumps` accepts.

        Raises:
            FeedMetadataNotFoundError
            StorageError

        .. versionadded:: 1.18
            Renamed from :meth:`get_feed_metadata`.

        """
        return zero_or_one(
            (v for _, v in self.get_feed_metadata(feed, key=key)),
            lambda: FeedMetadataNotFoundError(_feed_argument(feed), key),
            default,
        )

    def set_feed_metadata_item(
        self, feed: FeedInput, key: str, value: JSONType
    ) -> None:
        """Set metadata for a feed.

        Args:
            feed (str or Feed): The feed URL.
            key (str): The key of the metadata item to set.
            value (JSONType): The value of the metadata item to set.
                JSONType is whatever :py:func:`json.dumps` accepts.

        Raises:
            FeedNotFoundError
            StorageError

        .. versionadded:: 1.18
            Renamed from :meth:`set_feed_metadata`.

        """
        feed_url = _feed_argument(feed)
        self._storage.set_metadata((feed_url,), key, value)

    def delete_feed_metadata_item(self, feed: FeedInput, key: str) -> None:
        """Delete metadata for a feed.

        Args:
            feed (str or Feed): The feed URL.
            key (str): The key of the metadata item to delete.

        Raises:
            FeedMetadataNotFoundError
            StorageError

        .. versionadded:: 1.18
            Renamed from :meth:`delete_feed_metadata`.

        """
        feed_url = _feed_argument(feed)
        self._storage.delete_metadata((feed_url,), key)

    def enable_search(self) -> None:
        """Enable full-text search.

        Calling this method if search is already enabled is a no-op.

        Raises:
            SearchError
            StorageError

        """
        return self._search.enable()

    def disable_search(self) -> None:
        """Disable full-text search.

        Calling this method if search is already disabled is a no-op.

        Raises:
            SearchError

        """
        return self._search.disable()

    def is_search_enabled(self) -> bool:
        """Check if full-text search is enabled.

        Returns:
            bool: Whether search is enabled or not.

        Raises:
            SearchError

        """
        return self._search.is_enabled()

    def update_search(self) -> None:
        """Update the full-text search index.

        Search must be enabled to call this method.

        Raises:
            SearchNotEnabledError
            SearchError
            StorageError

        """
        return self._search.update()

    def search_entries(
        self,
        query: str,
        *,
        feed: Optional[FeedInput] = None,
        entry: Optional[EntryInput] = None,
        read: Optional[bool] = None,
        important: Optional[bool] = None,
        has_enclosures: Optional[bool] = None,
        feed_tags: TagFilterInput = None,
        sort: SearchSortOrder = 'relevant',
        limit: Optional[int] = None,
        starting_after: Optional[EntryInput] = None,
    ) -> Iterable[EntrySearchResult]:
        """Get entries matching a full-text search query.

        Entries are sorted according to ``sort``. Possible values:

        ``'relevant'``

            Most relevant first.

        ``'recent'``

            Most recent first. See :meth:`~Reader.get_entries()`
            for details on what *recent* means.

            .. versionadded:: 1.4

        ``'random'``

            Random order (shuffled). At at most 256 entries will be returned.

            .. versionadded:: 1.10

        Note:
            The query syntax is dependent on the search provider.

            The default (and for now, only) search provider is SQLite FTS5.
            You can find more details on its query syntax here:
            https://www.sqlite.org/fts5.html#full_text_query_syntax

            The columns available in queries are:

            * ``title``: the entry title
            * ``feed``: the feed title
            * ``content``: the entry main text content;
              this includes the summary and the value of contents that have
              text/(x)html, text/plain or missing content types

            Query examples:

            * ``hello internet``: entries that match "hello" and "internet"
            * ``hello NOT internet``: entries that match "hello" but do not
              match "internet"
            * ``hello feed: cortex``: entries that match "hello" anywhere,
              and their feed title matches "cortex"
            * ``hello NOT feed: internet``: entries that match "hello" anywhere,
              and their feed title does not match "internet"

        Search must be enabled to call this method.

        Args:
            query (str): The search query.
            feed (str or Feed or None): Only search the entries for this feed.
            entry (tuple(str, str) or Entry or None):
                Only search for the entry with this (feed URL, entry id) tuple.
            read (bool or None): Only search (un)read entries.
            important (bool or None): Only search (un)important entries.
            has_enclosures (bool or None): Only search entries that (don't)
                have enclosures.
            feed_tags (None or bool or list(str or bool or list(str or bool))):
                Only return the entries from feeds matching these tags;
                works like the :meth:`~Reader.get_feeds()` ``tags`` argument.
            sort (str): How to order results; one of ``'relevant'`` (default),
                ``'recent'``, or ``'random'``.
            limit (int or None): A limit on the number of results to be returned;
                by default, all results are returned.
            starting_after (tuple(str, str) or EntrySearchResult or None):
                Return results after this result; a cursor for use in pagination.
                Using ``starting_after`` with ``sort='random'`` is not supported.

        Yields:
            :class:`EntrySearchResult`: Sorted according to ``sort``.

        Raises:
            SearchNotEnabledError
            InvalidSearchQueryError
            SearchError
            StorageError
            EntryNotFoundError: If ``starting_after`` does not exist.

        .. versionadded:: 1.4
            The ``sort`` keyword argument.

        .. versionadded:: 1.7
            The ``feed_tags`` keyword argument.

        .. versionadded:: 1.12
            The ``limit`` and ``starting_after`` keyword arguments.

        """
        filter_options = EntryFilterOptions.from_args(
            feed, entry, read, important, has_enclosures, feed_tags
        )

        if sort not in ('relevant', 'recent', 'random'):
            raise ValueError("sort should be one of ('relevant', 'recent', 'random')")

        if limit is not None:
            if not isinstance(limit, numbers.Integral) or limit < 1:
                raise ValueError("limit should be a positive integer")

        if starting_after and sort == 'random':
            raise ValueError("using starting_after with sort='random' not supported")

        now = self._now()
        return self._search.search_entries(
            query,
            now,
            filter_options,
            sort,
            limit,
            _entry_argument(starting_after) if starting_after else None,
        )

    def search_entry_counts(
        self,
        query: str,
        *,
        feed: Optional[FeedInput] = None,
        entry: Optional[EntryInput] = None,
        read: Optional[bool] = None,
        important: Optional[bool] = None,
        has_enclosures: Optional[bool] = None,
        feed_tags: TagFilterInput = None,
    ) -> EntrySearchCounts:
        """Count entries matching a full-text search query.

        See :meth:`~Reader.search_entries()` for details on how
        the query syntax and filtering work.

        Search must be enabled to call this method.

        Args:
            query (str): The search query.
            feed (str or Feed or None): Only count the entries for this feed.
            entry (tuple(str, str) or Entry or None):
                Only count the entry with this (feed URL, entry id) tuple.
            read (bool or None): Only count (un)read entries.
            important (bool or None): Only count (un)important entries.
            has_enclosures (bool or None): Only count entries that (don't)
                have enclosures.
            feed_tags (None or bool or list(str or bool or list(str or bool))):
                Only count the entries from feeds matching these tags.

        Returns:
            EntrySearchCounts:

        Raises:
            SearchNotEnabledError
            InvalidSearchQueryError
            SearchError
            StorageError

        .. versionadded:: 1.11

        """

        filter_options = EntryFilterOptions.from_args(
            feed, entry, read, important, has_enclosures, feed_tags
        )
        now = self._now()
        return self._search.search_entry_counts(query, now, filter_options)

    def add_feed_tag(self, feed: FeedInput, tag: str) -> None:
        """Add a tag to a feed.

        Adding a tag that the feed already has is a no-op.

        Args:
            feed (str or Feed): The feed URL.
            tag (str): The tag to add.

        Raises:
            FeedNotFoundError
            StorageError

        .. versionadded:: 1.7

        """
        feed_url = _feed_argument(feed)
        self._storage.add_tag((feed_url,), tag)

    def remove_feed_tag(self, feed: FeedInput, tag: str) -> None:
        """Remove a tag from a feed.

        Removing a tag that the feed does not have is a no-op.

        Args:
            feed (str or Feed): The feed URL.
            tag (str): The tag to remove.

        Raises:
            StorageError

        .. versionadded:: 1.7

        """
        feed_url = _feed_argument(feed)
        self._storage.remove_tag((feed_url,), tag)

    def get_feed_tags(self, feed: Optional[FeedInput] = None) -> Iterable[str]:
        """Get all or some of the feed tags.

        Args:
            feed (str or Feed or None): Only return the tags for this feed.

        Yields:
            str: The tags, in alphabetical order.

        Raises:
            StorageError

        .. versionadded:: 1.7

        """
        feed_url = _feed_argument(feed) if feed is not None else None
        return self._storage.get_tags((feed_url,))

    def make_reader_reserved_name(self, key: str) -> str:
        """Create a *reader*-reserved tag or metadata name.
        See :ref:`reserved names` for details.

        Uses :attr:`~Reader.reserved_name_scheme` to build names of the format::

        {reader_prefix}{key}

        Using the default scheme:

        >>> reader.make_reader_reserved_name('key')
        '.reader.key'

        Args:
            key (str): A key.

        Returns:
            str: The name.

        .. versionadded:: 1.17

        """
        return self._reserved_name_scheme.make_reader_name(key)

    def make_plugin_reserved_name(
        self, plugin_name: str, key: Optional[str] = None
    ) -> str:
        """Create a plugin-reserved tag or metadata name.
        See :ref:`reserved names` for details.

        Plugins should use this to generate names
        for plugin-specific tags and metadata.

        Uses :attr:`~Reader.reserved_name_scheme` to build names of the format::

        {plugin_prefix}{plugin_name}
        {plugin_prefix}{plugin_name}{separator}{key}

        Using the default scheme:

        >>> reader.make_plugin_reserved_name('myplugin')
        '.plugin.myplugin'
        >>> reader.make_plugin_reserved_name('myplugin', 'key')
        '.plugin.myplugin.key'

        Args:
            plugin_name (str): The plugin package/module name.
            key (str or None): A key; if more than one reserved name is needed.

        Returns:
            str: The name.

        .. versionadded:: 1.17

        """
        return self._reserved_name_scheme.make_plugin_name(plugin_name, key)

    # Ideally, the getter would return a TypedDict,
    # but the setter would take *any* Mapping[str, str];
    # unfortunately, mypy doesn't like when the types differ:
    # https://github.com/python/mypy/issues/3004

    @property
    def reserved_name_scheme(self) -> Mapping[str, str]:
        """dict(str, str): Mapping used to build reserved names.
        See :meth:`~Reader.make_reader_reserved_name`
        and :meth:`~Reader.make_plugin_reserved_name`
        for details on how this is used.

        The default scheme (these keys are required)::

            {'reader_prefix': '.reader.', 'plugin_prefix': '.plugin.', 'separator': '.'}

        The returned mapping is immutable; assign a new mapping to change the scheme.

        .. versionadded:: 1.17

        """
        return MappingProxyType(self._reserved_name_scheme.__dict__)

    @reserved_name_scheme.setter
    def reserved_name_scheme(self, value: Mapping[str, str]) -> None:
        try:
            self._reserved_name_scheme = NameScheme.from_value(value)
        except Exception as e:
            raise AttributeError(f"invalid scheme: {value}") from e



import json
import os

import pytest
from fakeparser import Parser
from utils import utc_datetime as datetime


def test_regex_mark_as_read(make_reader):
    key = '.reader.mark_as_read'
    value = {'title': ['^match']}

    reader = make_reader(':memory:', plugins=['reader.mark_as_read'])

    parser = Parser()
    reader._parser = parser

    one = parser.feed(1, datetime(2010, 1, 1))
    parser.entry(1, 1, datetime(2010, 1, 1), title='match old')

    reader.add_feed(one)
    reader.update_feeds()

    reader.set_feed_metadata_item(one, key, value)

    one = parser.feed(1, datetime(2010, 1, 2))
    match_new = parser.entry(1, 2, datetime(2010, 1, 2), title='match new')
    parser.entry(1, 3, datetime(2010, 1, 2), title='no match')

    two = parser.feed(2, datetime(2010, 1, 2))
    parser.entry(2, 3, datetime(2010, 1, 2), title='match other')

    reader.add_feed(two)
    reader.update_feeds()

    assert len(list(reader.get_entries())) == 4
    assert set((e.id, e.read) for e in reader.get_entries(read=True)) == {
        (match_new.id, True),
    }

    parser.entry(1, 3, datetime(2010, 1, 2), title='no match once again')
    reader.update_feeds()

    assert set((e.id, e.read) for e in reader.get_entries(read=True)) == {
        (match_new.id, True),
    }


@pytest.mark.parametrize('value', ['x', {'title': 'x'}, {'title': [1]}])
def test_regex_mark_as_read_bad_metadata(make_reader, value):
    reader = make_reader(':memory:', plugins=['reader.mark_as_read'])

    parser = Parser()
    reader._parser = parser

    one = parser.feed(1, datetime(2010, 1, 1))
    parser.entry(1, 1, datetime(2010, 1, 1), title='match')

    reader.add_feed(one)
    reader.set_feed_metadata_item(one, '.reader.mark_as_read', value)

    reader.update_feeds()

    assert [e.read for e in reader.get_entries()] == [False]
